from collections.abc import Mapping
from typing import Any, TypeVar, cast

from attrs import define as _attrs_define
from attrs import field as _attrs_field

T = TypeVar("T", bound="StringData")


@_attrs_define
class StringData:
    """
    Attributes:
        input_strings (list[str]):
    """

    input_strings: list[str]
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        input_strings = self.input_strings

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({"input_strings": input_strings})

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        d = dict(src_dict)
        input_strings = cast(list[str], d.pop("input_strings"))

        string_data = cls(input_strings=input_strings)

        string_data.additional_properties = d
        return string_data

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from http import HTTPStatus
from typing import Any, Optional, Union

import httpx

from ... import errors
from ...client import AuthenticatedClient, Client
from ...models.base_prompt_template_version_response import BasePromptTemplateVersionResponse
from ...models.http_validation_error import HTTPValidationError
from ...types import UNSET, Response, Unset


def _get_kwargs(project_id: str, *, template_name: str, version: Union[None, Unset, int] = UNSET) -> dict[str, Any]:
    params: dict[str, Any] = {}

    params["template_name"] = template_name

    json_version: Union[None, Unset, int]
    if isinstance(version, Unset):
        json_version = UNSET
    else:
        json_version = version
    params["version"] = json_version

    params = {k: v for k, v in params.items() if v is not UNSET and v is not None}

    _kwargs: dict[str, Any] = {"method": "get", "url": f"/projects/{project_id}/templates/versions", "params": params}

    return _kwargs


def _parse_response(
    *, client: Union[AuthenticatedClient, Client], response: httpx.Response
) -> Optional[Union[BasePromptTemplateVersionResponse, HTTPValidationError]]:
    if response.status_code == 200:
        response_200 = BasePromptTemplateVersionResponse.from_dict(response.json())

        return response_200
    if response.status_code == 422:
        response_422 = HTTPValidationError.from_dict(response.json())

        return response_422
    if client.raise_on_unexpected_status:
        raise errors.UnexpectedStatus(response.status_code, response.content)
    else:
        return None


def _build_response(
    *, client: Union[AuthenticatedClient, Client], response: httpx.Response
) -> Response[Union[BasePromptTemplateVersionResponse, HTTPValidationError]]:
    return Response(
        status_code=HTTPStatus(response.status_code),
        content=response.content,
        headers=response.headers,
        parsed=_parse_response(client=client, response=response),
    )


def sync_detailed(
    project_id: str, *, client: AuthenticatedClient, template_name: str, version: Union[None, Unset, int] = UNSET
) -> Response[Union[BasePromptTemplateVersionResponse, HTTPValidationError]]:
    """Get Template Version By Name

     Get a prompt template from a project.

    Parameters
    ----------
    project_id : UUID4
        Prokect ID.
    template_name : str
        Prompt template name.
    version : Optional[int]
        Version number to fetch. defaults to selected version.
    ctx : Context, optional
        User context with database session, by default Depends(get_user_context).


    Returns
    -------
    GetTemplateResponse
        Prompt template response.

    Args:
        project_id (str):
        template_name (str):
        version (Union[None, Unset, int]):

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Response[Union[BasePromptTemplateVersionResponse, HTTPValidationError]]
    """

    kwargs = _get_kwargs(project_id=project_id, template_name=template_name, version=version)

    response = client.get_httpx_client().request(**kwargs)

    return _build_response(client=client, response=response)


def sync(
    project_id: str, *, client: AuthenticatedClient, template_name: str, version: Union[None, Unset, int] = UNSET
) -> Optional[Union[BasePromptTemplateVersionResponse, HTTPValidationError]]:
    """Get Template Version By Name

     Get a prompt template from a project.

    Parameters
    ----------
    project_id : UUID4
        Prokect ID.
    template_name : str
        Prompt template name.
    version : Optional[int]
        Version number to fetch. defaults to selected version.
    ctx : Context, optional
        User context with database session, by default Depends(get_user_context).


    Returns
    -------
    GetTemplateResponse
        Prompt template response.

    Args:
        project_id (str):
        template_name (str):
        version (Union[None, Unset, int]):

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Union[BasePromptTemplateVersionResponse, HTTPValidationError]
    """

    return sync_detailed(project_id=project_id, client=client, template_name=template_name, version=version).parsed


async def asyncio_detailed(
    project_id: str, *, client: AuthenticatedClient, template_name: str, version: Union[None, Unset, int] = UNSET
) -> Response[Union[BasePromptTemplateVersionResponse, HTTPValidationError]]:
    """Get Template Version By Name

     Get a prompt template from a project.

    Parameters
    ----------
    project_id : UUID4
        Prokect ID.
    template_name : str
        Prompt template name.
    version : Optional[int]
        Version number to fetch. defaults to selected version.
    ctx : Context, optional
        User context with database session, by default Depends(get_user_context).


    Returns
    -------
    GetTemplateResponse
        Prompt template response.

    Args:
        project_id (str):
        template_name (str):
        version (Union[None, Unset, int]):

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Response[Union[BasePromptTemplateVersionResponse, HTTPValidationError]]
    """

    kwargs = _get_kwargs(project_id=project_id, template_name=template_name, version=version)

    response = await client.get_async_httpx_client().request(**kwargs)

    return _build_response(client=client, response=response)


async def asyncio(
    project_id: str, *, client: AuthenticatedClient, template_name: str, version: Union[None, Unset, int] = UNSET
) -> Optional[Union[BasePromptTemplateVersionResponse, HTTPValidationError]]:
    """Get Template Version By Name

     Get a prompt template from a project.

    Parameters
    ----------
    project_id : UUID4
        Prokect ID.
    template_name : str
        Prompt template name.
    version : Optional[int]
        Version number to fetch. defaults to selected version.
    ctx : Context, optional
        User context with database session, by default Depends(get_user_context).


    Returns
    -------
    GetTemplateResponse
        Prompt template response.

    Args:
        project_id (str):
        template_name (str):
        version (Union[None, Unset, int]):

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Union[BasePromptTemplateVersionResponse, HTTPValidationError]
    """

    return (
        await asyncio_detailed(project_id=project_id, client=client, template_name=template_name, version=version)
    ).parsed



from collections.abc import Mapping
from typing import Any, TypeVar, Union, cast

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..models.step_type import StepType

T = TypeVar("T", bound="AggregatedTraceViewNode")


@_attrs_define
class AggregatedTraceViewNode:
    """
    Attributes:
        has_children (bool):
        id (str):
        name (Union[None, str]):
        occurrences (int):
        parent_id (Union[None, str]):
        type_ (StepType):
    """

    has_children: bool
    id: str
    name: Union[None, str]
    occurrences: int
    parent_id: Union[None, str]
    type_: StepType
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        has_children = self.has_children

        id = self.id

        name: Union[None, str]
        name = self.name

        occurrences = self.occurrences

        parent_id: Union[None, str]
        parent_id = self.parent_id

        type_ = self.type_.value

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "has_children": has_children,
                "id": id,
                "name": name,
                "occurrences": occurrences,
                "parent_id": parent_id,
                "type": type_,
            }
        )

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        d = dict(src_dict)
        has_children = d.pop("has_children")

        id = d.pop("id")

        def _parse_name(data: object) -> Union[None, str]:
            if data is None:
                return data
            return cast(Union[None, str], data)

        name = _parse_name(d.pop("name"))

        occurrences = d.pop("occurrences")

        def _parse_parent_id(data: object) -> Union[None, str]:
            if data is None:
                return data
            return cast(Union[None, str], data)

        parent_id = _parse_parent_id(d.pop("parent_id"))

        type_ = StepType(d.pop("type"))

        aggregated_trace_view_node = cls(
            has_children=has_children, id=id, name=name, occurrences=occurrences, parent_id=parent_id, type_=type_
        )

        aggregated_trace_view_node.additional_properties = d
        return aggregated_trace_view_node

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



import builtins
import logging
import warnings
from typing import Optional, Union, overload

from galileo import Message
from galileo.base import BaseClientModel
from galileo.projects import Projects
from galileo.resources.api.prompts import (
    create_global_prompt_template_templates_post,
    create_prompt_template_with_version_projects_project_id_templates_post,
    delete_global_template_templates_template_id_delete,
    get_global_template_templates_template_id_get,
    get_global_template_version_templates_template_id_versions_version_get,
    get_project_templates_projects_project_id_templates_get,
    get_template_from_project_projects_project_id_templates_template_id_get,
    query_templates_templates_query_post,
)
from galileo.resources.models import (
    BasePromptTemplateResponse,
    BasePromptTemplateVersionResponse,
    CreatePromptTemplateWithVersionRequestBody,
    HTTPValidationError,
    ListPromptTemplateParams,
    PromptTemplateNameFilter,
    PromptTemplateNameFilterOperator,
)
from galileo.resources.types import Unset
from galileo.utils.exceptions import APIException

_logger = logging.getLogger(__name__)


class PromptTemplateAPIException(APIException):
    pass


class PromptTemplate(BasePromptTemplateResponse):
    def __init__(self, prompt_template: Union[None, BasePromptTemplateResponse] = None):
        """
        Initialize a PromptTemplate instance.

        Parameters
        ----------
        prompt_template : Union[None, BasePromptTemplateResponse], optional
            The prompt template data to initialize from. If None, creates an empty prompt template instance.
            Defaults to None.
        """
        if prompt_template is not None:
            super().__init__(
                all_available_versions=prompt_template.all_available_versions,
                created_at=prompt_template.created_at,
                creator=prompt_template.creator,
                id=prompt_template.id,
                max_version=prompt_template.max_version,
                name=prompt_template.name,
                selected_version=prompt_template.selected_version,
                selected_version_id=prompt_template.selected_version_id,
                template=prompt_template.template,
                total_versions=prompt_template.total_versions,
                updated_at=prompt_template.updated_at,
                all_versions=prompt_template.all_versions,
                permissions=prompt_template.permissions,
            )
            self.additional_properties = prompt_template.additional_properties.copy()


class PromptTemplateVersion(BasePromptTemplateVersionResponse):
    def __init__(self, prompt_template_version: Union[None, BasePromptTemplateVersionResponse] = None):
        if prompt_template_version is not None:
            super().__init__(
                id=prompt_template_version.id,
                lines_added=prompt_template_version.lines_added,
                lines_edited=prompt_template_version.lines_edited,
                lines_removed=prompt_template_version.lines_removed,
                model_changed=prompt_template_version.model_changed,
                settings=prompt_template_version.settings,
                settings_changed=prompt_template_version.settings_changed,
                template=prompt_template_version.template,
                version=prompt_template_version.version,
                output_type=prompt_template_version.output_type,
                raw=prompt_template_version.raw,
                created_at=prompt_template_version.created_at,
                updated_at=prompt_template_version.updated_at,
            )
            self.additional_properties = prompt_template_version.additional_properties.copy()


class PromptTemplates(BaseClientModel):
    def list(self, project_name: str) -> list[PromptTemplate]:
        project = Projects().get(name=project_name)
        if not project:
            raise ValueError(f"Project {project_name} does not exist")

        templates = get_project_templates_projects_project_id_templates_get.sync(
            # TODO: remove type ignore, when migrated to proper AuthenticatedClient
            project_id=project.id,
            client=self.client,  # type: ignore[arg-type]
        )

        if not templates or isinstance(templates, HTTPValidationError):
            return []

        return [PromptTemplate(prompt_template=prompt_template) for prompt_template in templates]

    def get(self, *, project_name: str, template_id: str) -> Optional[PromptTemplate]:
        project = Projects().get(name=project_name)
        if not project:
            raise ValueError(f"Project {project_name} does not exist")

        _logger.debug(f"Get template {template_id} from project {project.id}")
        template = get_template_from_project_projects_project_id_templates_template_id_get.sync(
            # TODO: remove type ignore, when migrated to proper AuthenticatedClient
            project_id=project.id,
            template_id=template_id,
            client=self.client,  # type: ignore[arg-type]
        )

        if not template or isinstance(template, HTTPValidationError):
            return None

        return PromptTemplate(prompt_template=template)

    def create(
        self, name: str, project_name: str, template: Union[builtins.list[Message], str]
    ) -> Optional[PromptTemplate]:
        project = Projects().get(name=project_name)
        if not project:
            raise ValueError(f"Project {project_name} does not exist")

        body = CreatePromptTemplateWithVersionRequestBody(name=name, template=template)  # type: ignore[arg-type]

        _logger.debug(f"{body}")
        response = create_prompt_template_with_version_projects_project_id_templates_post.sync_detailed(
            project_id=project.id,
            client=self.client,  # type: ignore[arg-type]
            body=body,  # type: ignore[arg-type]
        )

        if response.status_code != 200:
            raise PromptTemplateAPIException(response.content.decode("utf-8"))

        if not response.parsed or isinstance(response.parsed, HTTPValidationError):
            _logger.error(response)
            raise PromptTemplateAPIException(response.content.decode("utf-8"))

        return PromptTemplate(prompt_template=response.parsed)


class GlobalPromptTemplates(BaseClientModel):
    def list(
        self, *, name_filter: Optional[str] = None, limit: Union[Unset, int] = 100, starting_token: int = 0
    ) -> list[PromptTemplate]:
        params = ListPromptTemplateParams()
        if name_filter:
            params.filters = [
                PromptTemplateNameFilter(operator=PromptTemplateNameFilterOperator.CONTAINS, value=name_filter)
            ]

        response = query_templates_templates_query_post.sync(
            client=self.client, body=params, limit=limit, starting_token=starting_token
        )

        if not response or isinstance(response, HTTPValidationError):
            return []

        if hasattr(response, "templates") and response.templates:
            return [PromptTemplate(prompt_template=template) for template in response.templates]

        return []

    @overload
    def get(self, *, template_id: str) -> Optional[PromptTemplate]: ...

    @overload
    def get(self, *, name: str) -> Optional[PromptTemplate]: ...

    def get(self, *, template_id: Optional[str] = None, name: Optional[str] = None) -> Optional[PromptTemplate]:
        if (template_id is None) and (name is None):
            raise ValueError("Exactly one of 'template_id' or 'name' must be provided")

        if (template_id is not None) and (name is not None):
            raise ValueError("Exactly one of 'template_id' or 'name' must be provided")

        if template_id:
            _logger.debug(f"Get global template {template_id}")
            template = get_global_template_templates_template_id_get.sync(template_id=template_id, client=self.client)

            if not template or isinstance(template, HTTPValidationError):
                return None

            return PromptTemplate(prompt_template=template)

        if name:
            templates = self.list(name_filter=name, limit=1)
            for template in templates:
                if template.name == name:
                    return template
            return None

        return None

    @overload
    def delete(self, *, template_id: str) -> None: ...

    @overload
    def delete(self, *, name: str) -> None: ...

    def delete(self, *, template_id: Optional[str] = None, name: Optional[str] = None) -> None:
        if (template_id is None) == (name is None):
            raise ValueError("Exactly one of 'id' or 'name' must be provided")

        if name:
            template = self.get(name=name)
            if not template:
                raise ValueError(f"Global template '{name}' not found")
            template_id = template.id

        if template_id:
            delete_global_template_templates_template_id_delete.sync(client=self.client, template_id=template_id)

    def get_version(self, *, template_id: str, version: int) -> Optional[PromptTemplateVersion]:
        _logger.debug(f"Get global template {template_id} version {version}")
        template_version = get_global_template_version_templates_template_id_versions_version_get.sync(
            template_id=template_id, version=version, client=self.client
        )

        if not template_version or isinstance(template_version, HTTPValidationError):
            return None

        return PromptTemplateVersion(prompt_template_version=template_version)

    def create(self, name: str, template: Union[builtins.list[Message], str]) -> PromptTemplate:
        body = CreatePromptTemplateWithVersionRequestBody(name=name, template=template)

        _logger.debug(f"Creating global template: {body}")
        response = create_global_prompt_template_templates_post.sync_detailed(client=self.client, body=body)

        if response.status_code != 200:
            raise PromptTemplateAPIException(response.content.decode("utf-8"))

        if not response.parsed or isinstance(response.parsed, HTTPValidationError):
            _logger.error(response)
            raise PromptTemplateAPIException(response.content.decode("utf-8"))

        return PromptTemplate(prompt_template=response.parsed)


def create_prompt_template(name: str, project: str, messages: Union[list[Message], str]) -> Optional[PromptTemplate]:
    warnings.warn("create_prompt_template is deprecated, use create_prompt instead.", DeprecationWarning, stacklevel=2)
    return PromptTemplates().create(name=name, project_name=project, template=messages)


def get_prompt_template(project: str, name: str) -> Optional[PromptTemplate]:
    warnings.warn("get_prompt_template is deprecated, use get_prompt instead.", DeprecationWarning, stacklevel=2)
    prompt_templates = PromptTemplates().list(project_name=project)
    for prompt_template in prompt_templates:
        if prompt_template.name == name:
            _logger.info(f"Get template {prompt_template}")
            return prompt_template
    _logger.warning(f"Template {name} not found in project {project}")
    return None


def list_prompt_templates(project: str) -> list[PromptTemplate]:
    warnings.warn("list_prompt_templates is deprecated, use get_prompts instead.", DeprecationWarning, stacklevel=2)
    return PromptTemplates().list(project_name=project)


@overload
def get_prompt(*, id: str) -> Optional[PromptTemplateVersion]: ...


@overload
def get_prompt(*, name: str) -> Optional[PromptTemplateVersion]: ...


def get_prompt(
    *, id: Optional[str] = None, name: Optional[str] = None, version: Optional[int] = None
) -> Optional[PromptTemplateVersion]:
    """
    Retrieves a specific global prompt template version.

    You must provide either 'id' or 'name', but not both.
    If 'version' is not provided, the currently selected version is returned.
    Parameters
    ----------
    id : str, optional
        The unique identifier of the template to retrieve. Defaults to None.
    name : str, optional
        The name of the template to retrieve. Defaults to None.
    version : int, optional
        The version number to retrieve. If not provided, the currently selected version is returned. Defaults to None.
    Returns
    -------
    Optional[PromptTemplateVersion]
        The template version if found, None otherwise.
    Raises
    ------
    ValueError
        If neither or both 'id' and 'name' are provided.
    """
    if (id is None) and (name is None):
        raise ValueError("Exactly one of 'id' or 'name' must be provided")

    if (id is not None) and (name is not None):
        raise ValueError("Exactly one of 'id' or 'name' must be provided")

    if version is not None:
        target_id = id
        if name:
            prompt_template = GlobalPromptTemplates().get(name=name)
            if not prompt_template:
                return None
            target_id = prompt_template.id

        if not target_id:
            raise ValueError("A template id is required to fetch a specific version.")

        return GlobalPromptTemplates().get_version(template_id=target_id, version=version)

    prompt_template = GlobalPromptTemplates().get(template_id=id) if id else GlobalPromptTemplates().get(name=name)  # type: ignore[arg-type]

    if not prompt_template or isinstance(prompt_template.selected_version, Unset):
        return None

    return PromptTemplateVersion(prompt_template_version=prompt_template.selected_version)


@overload
def delete_prompt(*, id: str) -> None: ...


@overload
def delete_prompt(*, name: str) -> None: ...


def delete_prompt(*, id: Optional[str] = None, name: Optional[str] = None) -> None:
    """
    Delete a global prompt template by ID or name.

    Parameters
    ----------
    id : str
        The unique identifier of the template to delete. Defaults to None.
    name : str
        The name of the template to delete. Defaults to None.

    Returns
    -------
    None

    Raises
    ------
    ValueError
        If neither or both id and name are provided, or if the template is not found.
    """
    return GlobalPromptTemplates().delete(template_id=id, name=name)  # type: ignore[call-overload]


def create_prompt(name: str, template: Union[list[Message], str]) -> PromptTemplate:
    """
    Create a new global prompt template.

    Parameters
    ----------
    name : str
        The name for the new template.
    template : Union[list[Message], str]
        The template content. Can be either a list of Message objects or a JSON string
        representing the message structure.

    Returns
    -------
    PromptTemplate
        The created prompt template.

    Raises
    ------
    PromptTemplateAPIException
        If the API request fails or returns an error.
    """
    return GlobalPromptTemplates().create(name=name, template=template)


def get_prompts(name_filter: Optional[str] = None, limit: Union[Unset, int] = 100) -> list[PromptTemplate]:
    """
    List global prompt templates with optional filtering.

    Parameters
    ----------
    name_filter : Optional[str], optional
        Filter templates by name containing this string. Defaults to None (no filtering).
    limit : Union[Unset, int], optional
        Maximum number of templates to return. Defaults to 100.

    Returns
    -------
    list[PromptTemplate]
        List of prompt templates matching the criteria.

    Examples
    --------
    >>> # List all global templates
    >>> templates = get_prompts()

    >>> # List templates with names containing "assistant"
    >>> templates = get_prompts(name_filter="assistant")

    >>> # List first 10 templates
    >>> templates = get_prompts(limit=10)
    """
    return GlobalPromptTemplates().list(name_filter=name_filter, limit=limit)



from collections.abc import Mapping
from typing import TYPE_CHECKING, Any, Literal, TypeVar, Union, cast

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.recompute_settings_log_stream import RecomputeSettingsLogStream
    from ..models.recompute_settings_observe import RecomputeSettingsObserve
    from ..models.recompute_settings_project import RecomputeSettingsProject
    from ..models.recompute_settings_runs import RecomputeSettingsRuns


T = TypeVar("T", bound="MetricCritiqueJobConfiguration")


@_attrs_define
class MetricCritiqueJobConfiguration:
    """Info necessary to execute a metric critique job.

    Attributes:
        critique_ids (list[str]):
        metric_name (str):
        project_type (Union[Literal['gen_ai'], Literal['llm_monitor'], Literal['prompt_evaluation']]):
        recompute_settings (Union['RecomputeSettingsLogStream', 'RecomputeSettingsObserve', 'RecomputeSettingsProject',
            'RecomputeSettingsRuns', None, Unset]):
        scorer_id (Union[None, Unset, str]):
    """

    critique_ids: list[str]
    metric_name: str
    project_type: Union[Literal["gen_ai"], Literal["llm_monitor"], Literal["prompt_evaluation"]]
    recompute_settings: Union[
        "RecomputeSettingsLogStream",
        "RecomputeSettingsObserve",
        "RecomputeSettingsProject",
        "RecomputeSettingsRuns",
        None,
        Unset,
    ] = UNSET
    scorer_id: Union[None, Unset, str] = UNSET
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        from ..models.recompute_settings_log_stream import RecomputeSettingsLogStream
        from ..models.recompute_settings_observe import RecomputeSettingsObserve
        from ..models.recompute_settings_project import RecomputeSettingsProject
        from ..models.recompute_settings_runs import RecomputeSettingsRuns

        critique_ids = self.critique_ids

        metric_name = self.metric_name

        project_type: Union[Literal["gen_ai"], Literal["llm_monitor"], Literal["prompt_evaluation"]]
        project_type = self.project_type

        recompute_settings: Union[None, Unset, dict[str, Any]]
        if isinstance(self.recompute_settings, Unset):
            recompute_settings = UNSET
        elif isinstance(self.recompute_settings, RecomputeSettingsRuns):
            recompute_settings = self.recompute_settings.to_dict()
        elif isinstance(self.recompute_settings, RecomputeSettingsProject):
            recompute_settings = self.recompute_settings.to_dict()
        elif isinstance(self.recompute_settings, RecomputeSettingsObserve):
            recompute_settings = self.recompute_settings.to_dict()
        elif isinstance(self.recompute_settings, RecomputeSettingsLogStream):
            recompute_settings = self.recompute_settings.to_dict()
        else:
            recompute_settings = self.recompute_settings

        scorer_id: Union[None, Unset, str]
        if isinstance(self.scorer_id, Unset):
            scorer_id = UNSET
        else:
            scorer_id = self.scorer_id

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({"critique_ids": critique_ids, "metric_name": metric_name, "project_type": project_type})
        if recompute_settings is not UNSET:
            field_dict["recompute_settings"] = recompute_settings
        if scorer_id is not UNSET:
            field_dict["scorer_id"] = scorer_id

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        from ..models.recompute_settings_log_stream import RecomputeSettingsLogStream
        from ..models.recompute_settings_observe import RecomputeSettingsObserve
        from ..models.recompute_settings_project import RecomputeSettingsProject
        from ..models.recompute_settings_runs import RecomputeSettingsRuns

        d = dict(src_dict)
        critique_ids = cast(list[str], d.pop("critique_ids"))

        metric_name = d.pop("metric_name")

        def _parse_project_type(
            data: object,
        ) -> Union[Literal["gen_ai"], Literal["llm_monitor"], Literal["prompt_evaluation"]]:
            project_type_type_0 = cast(Literal["prompt_evaluation"], data)
            if project_type_type_0 != "prompt_evaluation":
                raise ValueError(
                    f"project_type_type_0 must match const 'prompt_evaluation', got '{project_type_type_0}'"
                )
            return project_type_type_0
            project_type_type_1 = cast(Literal["llm_monitor"], data)
            if project_type_type_1 != "llm_monitor":
                raise ValueError(f"project_type_type_1 must match const 'llm_monitor', got '{project_type_type_1}'")
            return project_type_type_1
            project_type_type_2 = cast(Literal["gen_ai"], data)
            if project_type_type_2 != "gen_ai":
                raise ValueError(f"project_type_type_2 must match const 'gen_ai', got '{project_type_type_2}'")
            return project_type_type_2

        project_type = _parse_project_type(d.pop("project_type"))

        def _parse_recompute_settings(
            data: object,
        ) -> Union[
            "RecomputeSettingsLogStream",
            "RecomputeSettingsObserve",
            "RecomputeSettingsProject",
            "RecomputeSettingsRuns",
            None,
            Unset,
        ]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            try:
                if not isinstance(data, dict):
                    raise TypeError()
                recompute_settings_type_0_type_0 = RecomputeSettingsRuns.from_dict(data)

                return recompute_settings_type_0_type_0
            except:  # noqa: E722
                pass
            try:
                if not isinstance(data, dict):
                    raise TypeError()
                recompute_settings_type_0_type_1 = RecomputeSettingsProject.from_dict(data)

                return recompute_settings_type_0_type_1
            except:  # noqa: E722
                pass
            try:
                if not isinstance(data, dict):
                    raise TypeError()
                recompute_settings_type_0_type_2 = RecomputeSettingsObserve.from_dict(data)

                return recompute_settings_type_0_type_2
            except:  # noqa: E722
                pass
            try:
                if not isinstance(data, dict):
                    raise TypeError()
                recompute_settings_type_0_type_3 = RecomputeSettingsLogStream.from_dict(data)

                return recompute_settings_type_0_type_3
            except:  # noqa: E722
                pass
            return cast(
                Union[
                    "RecomputeSettingsLogStream",
                    "RecomputeSettingsObserve",
                    "RecomputeSettingsProject",
                    "RecomputeSettingsRuns",
                    None,
                    Unset,
                ],
                data,
            )

        recompute_settings = _parse_recompute_settings(d.pop("recompute_settings", UNSET))

        def _parse_scorer_id(data: object) -> Union[None, Unset, str]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, str], data)

        scorer_id = _parse_scorer_id(d.pop("scorer_id", UNSET))

        metric_critique_job_configuration = cls(
            critique_ids=critique_ids,
            metric_name=metric_name,
            project_type=project_type,
            recompute_settings=recompute_settings,
            scorer_id=scorer_id,
        )

        metric_critique_job_configuration.additional_properties = d
        return metric_critique_job_configuration

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



import builtins
from typing import Optional, Union, overload

from galileo.base import BaseClientModel
from galileo.projects import Projects
from galileo.resources.api.log_stream import (
    create_log_stream_projects_project_id_log_streams_post,
    get_log_stream_projects_project_id_log_streams_log_stream_id_get,
    list_log_streams_projects_project_id_log_streams_get,
)
from galileo.resources.models.http_validation_error import HTTPValidationError
from galileo.resources.models.log_stream_create_request import LogStreamCreateRequest
from galileo.resources.models.log_stream_response import LogStreamResponse
from galileo.utils.catch_log import DecorateAllMethods


class LogStream(LogStreamResponse):
    """
    Log streams are used to organize logs within a project on the Galileo platform.
    They provide a way to categorize and group related logs, making it easier to
    analyze and monitor specific parts of your application or different environments
    (e.g., production, staging, development).

    Attributes
    ----------
    created_at : datetime.datetime
        The timestamp when the log stream was created.
    created_by : str
        The identifier of the user who created the log stream.
    id : str
        The unique identifier of the log stream.
    name : str
        The name of the log stream.
    project_id : str
        The ID of the project this log stream belongs to.
    updated_at : datetime.datetime
        The timestamp when the log stream was last updated.
    additional_properties : dict
        Additional properties associated with the log stream.

    Examples
    --------
    # Create a new log stream in a project
    from galileo.log_streams import create_log_stream

    # Create by project ID
    log_stream = create_log_stream(name="Production Logs", project_id="project-123")

    # Create by project name
    log_stream = create_log_stream(name="Production Logs", project_name="My AI Project")

    # Get a log stream by name
    from galileo.log_streams import get_log_stream
    log_stream = get_log_stream(name="Production Logs", project_name="My AI Project")

    # List all log streams in a project
    from galileo.log_streams import list_log_streams
    log_streams = list_log_streams(project_name="My AI Project")
    for stream in log_streams:
        print(f"Log Stream: {stream.name} (ID: {stream.id})")

    # Use a log stream with the context manager
    from galileo.openai import openai
    from galileo import galileo_context

    with galileo_context(project="My AI Project", log_stream="Production Logs"):
        response = openai.chat.completions.create(
            model="gpt-4o",
            messages=[{"role": "user", "content": "Hello, world!"}]
        )
    """

    def __init__(self, log_stream: Union[None, LogStreamResponse] = None):
        """
        Initialize a LogStream instance.

        Parameters
        ----------
        log_stream : Union[None, LogStreamResponse], optional
            The log stream data to initialize from. If None, creates an empty log stream instance.
            Defaults to None.
        """
        if log_stream is not None:
            super().__init__(
                created_at=log_stream.created_at,
                id=log_stream.id,
                name=log_stream.name,
                project_id=log_stream.project_id,
                updated_at=log_stream.updated_at,
                created_by=log_stream.created_by,
            )
            self.additional_properties = log_stream.additional_properties.copy()
            return


class LogStreams(BaseClientModel, DecorateAllMethods):
    @overload
    def list(self, *, project_id: str) -> list[LogStream]: ...

    @overload
    def list(self, *, project_name: str) -> builtins.list[LogStream]: ...

    def list(self, *, project_id: Optional[str] = None, project_name: Optional[str] = None) -> builtins.list[LogStream]:
        """
        Lists all log streams. Exactly one of `project_id` or `project_name` must be provided.

        Parameters
        ----------
        project_id : Optional[str], optional
            The ID of the project to list log streams for.
        project_name : Optional[str], optional
            The name of the project to list log streams for.

        Returns
        -------
        builtins.list[LogStream]
            A list of log streams.

        Raises
        ------
        ValueError
            If neither or both `project_id` and `project_name` are provided.
        errors.UnexpectedStatus
            If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException
            If the request takes longer than Client.timeout.
        """
        if (project_id is None) and (project_name is None):
            raise ValueError("Exactly one of 'project_id' or 'project_name' must be provided")

        if project_id:
            log_streams = list_log_streams_projects_project_id_log_streams_get.sync(
                client=self.client, project_id=project_id
            )
        else:
            project = Projects(client=self.client).get(name=project_name)
            if not project:
                raise ValueError(f"Project {project_name} not found")
            log_streams = list_log_streams_projects_project_id_log_streams_get.sync(
                client=self.client, project_id=project.id
            )
        return [LogStream(log_stream=log_stream) for log_stream in log_streams] if log_streams else []

    @overload
    def get(
        self, *, id: str, project_id: Optional[str] = None, project_name: Optional[str] = None
    ) -> Optional[LogStream]: ...
    @overload
    def get(
        self, *, name: str, project_id: Optional[str] = None, project_name: Optional[str] = None
    ) -> Optional[LogStream]: ...
    def get(
        self,
        *,
        id: Optional[str] = None,
        name: Optional[str] = None,
        project_id: Optional[str] = None,
        project_name: Optional[str] = None,
    ) -> Optional[LogStream]:
        """
        Retrieves a log stream by id or name.

        Parameters
        ----------
        id : Optional[str], optional
            The id of the log stream. Defaults to None.
        name : Optional[str], optional
            The name of the log stream. Defaults to None.
        project_id : Optional[str], optional
            The ID of the project. Defaults to None.
        project_name : Optional[str], optional
            The name of the project. Defaults to None.

        Returns
        -------
        Optional[LogStream]
            The log stream if found, None otherwise.

        Raises
        ------
        ValueError
            If neither or both `id` and `name` are provided, or if neither or both
            `project_id` and `project_name` are provided.
        errors.UnexpectedStatus
            If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException
            If the request takes longer than Client.timeout.
        """
        if (id is None) == (name is None):
            raise ValueError("Exactly one of 'id' or 'name' must be provided")

        if (project_id is None) == (project_name is None):
            raise ValueError("Exactly one of 'project_id' or 'project_name' must be provided")

        if not project_id:
            project = Projects(client=self.client).get(name=project_name)
            if not project:
                raise ValueError(f"Project {project_name} not found")
            project_id = project.id

        if id:
            log_stream_response = get_log_stream_projects_project_id_log_streams_log_stream_id_get.sync(
                project_id=project_id, log_stream_id=id, client=self.client
            )
            if not log_stream_response:
                return None
            return LogStream(log_stream=log_stream_response)

        elif name:
            log_streams = self.list(project_id=project_id)

            if not log_streams or len(log_streams) == 0:
                return None

            for log_stream in log_streams:
                if log_stream.name == name:
                    return log_stream
        return None

    @overload
    def create(self, name: str, *, project_id: Optional[str] = None) -> LogStream: ...
    @overload
    def create(self, name: str, *, project_name: str) -> LogStream: ...

    def create(self, name: str, *, project_id: Optional[str] = None, project_name: Optional[str] = None) -> LogStream:
        """
        Creates a new log stream. Exactly one of `project_id` or `project_name` must be provided.

        Parameters
        ----------
        name : str
            The name of the log stream.
        project_id : Optional[str], optional
            The ID of the project to create the log stream in. Defaults to None.
        project_name : Optional[str], optional
            The name of the project to create the log stream in. Defaults to None.

        Returns
        -------
        LogStream
            The created log stream.

        Raises
        ------
        ValueError
            If neither or both `project_id` and `project_name` are provided, or if the project is not found.
        HTTPValidationError
            If the server validation fails.
        errors.UnexpectedStatus
            If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException
            If the request takes longer than Client.timeout.
        """

        if (project_id is None) == (project_name is None):
            raise ValueError("Exactly one of 'project_id' or 'project_name' must be provided")

        if not project_id:
            project = Projects(client=self.client).get(name=project_name)
            if not project:
                raise ValueError(f"Project {project_name} not found")
            project_id = project.id

        body = LogStreamCreateRequest(name=name)
        response = create_log_stream_projects_project_id_log_streams_post.sync(
            project_id=project_id, client=self.client, body=body
        )

        if isinstance(response, HTTPValidationError):
            raise response

        if not response:
            raise ValueError("Unable to create log stream")

        return LogStream(log_stream=response)


#
# Convenience methods
#


def get_log_stream(
    *, name: Optional[str] = None, project_id: Optional[str] = None, project_name: Optional[str] = None
) -> Optional[LogStream]:
    """
    Retrieves a log stream by name. Exactly one of `project_id` or `project_name` must be provided.

    Parameters
    ----------
    name : Optional[str], optional
        The name of the log stream. Defaults to None.
    project_id : Optional[str], optional
        The ID of the project. Defaults to None.
    project_name : Optional[str], optional
        The name of the project. Defaults to None.

    Returns
    -------
    Optional[LogStream]
        The log stream if found, None otherwise.

    Raises
    ------
    ValueError
        If neither or both `project_id` and `project_name` are provided.
    errors.UnexpectedStatus
        If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
    httpx.TimeoutException
        If the request takes longer than Client.timeout.
    """
    return LogStreams().get(name=name, project_id=project_id, project_name=project_name)


def list_log_streams(*, project_id: Optional[str] = None, project_name: Optional[str] = None) -> list[LogStream]:
    """
    Lists all log streams. Exactly one of `project_id` or `project_name` must be provided.

    Parameters
    ----------
    project_id : str
        The id of the project.
    project_name : str
        The name of the project.

    Returns
    -------
    list[LogStream]
        A list of projects.

    Raises
    ------
    errors.UnexpectedStatus
        If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
    httpx.TimeoutException
        If the request takes longer than Client.timeout.

    """
    return LogStreams().list(project_id=project_id, project_name=project_name)


def create_log_stream(name: str, project_id: Optional[str] = None, project_name: Optional[str] = None) -> LogStream:
    """
    Creates a new log stream. Exactly one of `project_id` or `project_name` must be provided.

    Parameters
    ----------
    name : str
        The name of the log stream.

    Returns
    -------
    LogStream
        The created project.

    Raises
    ------
    errors.UnexpectedStatus
        If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
    httpx.TimeoutException
        If the request takes longer than Client.timeout.

    """
    return LogStreams().create(name=name, project_id=project_id, project_name=project_name)



from collections.abc import Mapping
from typing import Any, Literal, TypeVar, Union, cast

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..models.scorer_type import ScorerType
from ..types import UNSET, Unset

T = TypeVar("T", bound="MetricFailed")


@_attrs_define
class MetricFailed:
    """
    Attributes:
        message (Union[None, Unset, str]):  Default: 'Metric failed to compute.'.
        scorer_type (Union[None, ScorerType, Unset]):
        status_type (Union[Literal['failed'], Unset]):  Default: 'failed'.
    """

    message: Union[None, Unset, str] = "Metric failed to compute."
    scorer_type: Union[None, ScorerType, Unset] = UNSET
    status_type: Union[Literal["failed"], Unset] = "failed"
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        message: Union[None, Unset, str]
        if isinstance(self.message, Unset):
            message = UNSET
        else:
            message = self.message

        scorer_type: Union[None, Unset, str]
        if isinstance(self.scorer_type, Unset):
            scorer_type = UNSET
        elif isinstance(self.scorer_type, ScorerType):
            scorer_type = self.scorer_type.value
        else:
            scorer_type = self.scorer_type

        status_type = self.status_type

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if message is not UNSET:
            field_dict["message"] = message
        if scorer_type is not UNSET:
            field_dict["scorer_type"] = scorer_type
        if status_type is not UNSET:
            field_dict["status_type"] = status_type

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        d = dict(src_dict)

        def _parse_message(data: object) -> Union[None, Unset, str]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, str], data)

        message = _parse_message(d.pop("message", UNSET))

        def _parse_scorer_type(data: object) -> Union[None, ScorerType, Unset]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            try:
                if not isinstance(data, str):
                    raise TypeError()
                scorer_type_type_0 = ScorerType(data)

                return scorer_type_type_0
            except:  # noqa: E722
                pass
            return cast(Union[None, ScorerType, Unset], data)

        scorer_type = _parse_scorer_type(d.pop("scorer_type", UNSET))

        status_type = cast(Union[Literal["failed"], Unset], d.pop("status_type", UNSET))
        if status_type != "failed" and not isinstance(status_type, Unset):
            raise ValueError(f"status_type must match const 'failed', got '{status_type}'")

        metric_failed = cls(message=message, scorer_type=scorer_type, status_type=status_type)

        metric_failed.additional_properties = d
        return metric_failed

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from enum import Enum


class ColumnCategory(str, Enum):
    DATASET = "dataset"
    DATASET_METADATA = "dataset_metadata"
    FEEDBACK = "feedback"
    METRIC = "metric"
    STANDARD = "standard"
    USER_METADATA = "user_metadata"

    def __str__(self) -> str:
        return str(self.value)



from typing import Optional, Union

from pydantic import UUID4

from galileo.api_client import GalileoApiClient
from galileo.base import BaseClientModel
from galileo.projects import Projects
from galileo.resources.api.protect import (
    create_stage_projects_project_id_stages_post,
    get_stage_projects_project_id_stages_get,
    pause_stage_projects_project_id_stages_stage_id_put,
    update_stage_projects_project_id_stages_stage_id_post,
)
from galileo.resources.models.rulesets_mixin import RulesetsMixin as APIRulesetsMixin
from galileo.resources.models.stage_db import StageDB as APIStageDB
from galileo.resources.models.stage_with_rulesets import StageWithRulesets as APIStageWithRulesets
from galileo.resources.types import UNSET
from galileo.utils.catch_log import DecorateAllMethods
from galileo_core.schemas.protect.rule import Rule
from galileo_core.schemas.protect.ruleset import Ruleset, RulesetsMixin
from galileo_core.schemas.protect.stage import StageDB, StageType, StageWithRulesets
from galileo_core.utils.name import ts_name


def _get_project_id(
    project_id: Optional[Union[str, UUID4]] = None,
    project_name: Optional[str] = None,
    client: Optional[GalileoApiClient] = None,
) -> Optional[str]:
    """
    Resolves project ID from either project_id or project_name.
    """
    if project_id:
        return str(project_id)
    elif project_name:
        project = Projects(client=client).get(name=project_name)
        if not project:
            raise ValueError(f"Project with name '{project_name}' not found.")
        return str(project.id)
    else:
        raise ValueError("Either project_id or project_name must be provided.")


def _get_stage_id(
    stage_id: Optional[Union[str, UUID4]] = None,
    stage_name: Optional[str] = None,
    project_id: Optional[Union[str, UUID4]] = None,
    client: Optional[GalileoApiClient] = None,
) -> Optional[str]:
    """
    Resolves stage ID from either stage_id or stage_name.
    If stage_name is provided, it will look up the stage within the specified project.
    """
    if stage_id:
        return str(stage_id)
    elif stage_name:
        stage = Stages(client=client).get(project_id=project_id, stage_name=stage_name)
        if not stage:
            raise ValueError(f"Stage with name '{stage_name}' not found.")
        return str(stage.id)
    else:
        raise ValueError("Either stage_id or stage_name must be provided.")


class Stages(BaseClientModel, DecorateAllMethods):
    def create(
        self,
        project_id: Union[str, UUID4],
        name: Optional[str] = None,
        stage_type: StageType = StageType.local,
        pause: bool = False,
        rulesets: Optional[list[Rule]] = None,
        description: Optional[str] = None,
    ) -> StageDB:
        if not project_id:
            raise ValueError("project_id must be provided for creating a stage.")
        actual_name = name or ts_name("stage")

        request = StageWithRulesets(
            name=actual_name,
            project_id=str(project_id),
            type=stage_type,
            paused=pause,
            description=description,
            prioritized_rulesets=[Ruleset(rules=rulesets)] if rulesets else [],
        )

        request_dict = request.model_dump(mode="json")
        request_dict["prioritized_rulesets"] = request_dict.pop("rulesets", [])
        body = APIStageWithRulesets.from_dict(request_dict)

        response = create_stage_projects_project_id_stages_post.sync(
            project_id=str(project_id), client=self.client, body=body
        )

        if isinstance(response, APIStageDB):
            return StageDB.model_validate(response.to_dict())
        return response

    def get(
        self,
        project_id: Optional[Union[str, UUID4]] = None,
        project_name: Optional[str] = None,
        stage_id: Optional[Union[str, UUID4]] = None,
        stage_name: Optional[str] = None,
    ) -> StageDB:
        actual_project_id: Optional[str] = _get_project_id(
            project_id=project_id, project_name=project_name, client=self.client
        )

        if not stage_id and not stage_name:
            raise ValueError("Either stage_id or stage_name must be provided.")

        response = get_stage_projects_project_id_stages_get.sync(
            project_id=actual_project_id,
            stage_id=str(stage_id) if stage_id else UNSET,
            stage_name=stage_name if stage_name else UNSET,
            client=self.client,
        )
        if isinstance(response, APIStageDB):
            return StageDB.model_validate(response.to_dict())
        return response

    def update(
        self,
        project_id: Optional[Union[str, UUID4]] = None,
        project_name: Optional[str] = None,
        stage_id: Optional[Union[str, UUID4]] = None,
        stage_name: Optional[str] = None,
        prioritized_rulesets: Optional[list[Rule]] = None,
    ) -> StageDB:
        actual_project_id: Optional[str] = _get_project_id(
            project_id=project_id, project_name=project_name, client=self.client
        )

        actual_stage_id: Optional[str] = _get_stage_id(
            stage_id=stage_id, stage_name=stage_name, project_id=actual_project_id, client=self.client
        )

        rulesets = [Ruleset(rules=prioritized_rulesets)] if prioritized_rulesets else []
        request = RulesetsMixin(prioritized_rulesets=rulesets)
        request_dict = request.model_dump()
        request_dict["prioritized_rulesets"] = request_dict.pop("rulesets", [])
        body = APIRulesetsMixin.from_dict(request_dict)

        response = update_stage_projects_project_id_stages_stage_id_post.sync(
            project_id=actual_project_id, stage_id=actual_stage_id, client=self.client, body=body
        )
        if isinstance(response, APIStageDB):
            return StageDB.model_validate(response.to_dict())
        return response

    def _set_pause_state(
        self,
        pause_flag: bool,
        project_id: Optional[Union[str, UUID4]] = None,
        project_name: Optional[str] = None,
        stage_id: Optional[Union[str, UUID4]] = None,
        stage_name: Optional[str] = None,
    ) -> StageDB:
        """Sets the pause state of a stage."""
        actual_project_id: Optional[str] = _get_project_id(
            project_id=project_id, project_name=project_name, client=self.client
        )

        actual_stage_id: Optional[str] = _get_stage_id(
            stage_id=stage_id, stage_name=stage_name, project_id=actual_project_id, client=self.client
        )

        response = pause_stage_projects_project_id_stages_stage_id_put.sync(
            project_id=actual_project_id, stage_id=actual_stage_id, client=self.client, pause=pause_flag
        )
        if isinstance(response, APIStageDB):
            return StageDB.model_validate(response.to_dict())
        return response

    def pause(
        self,
        project_id: Optional[Union[str, UUID4]] = None,
        project_name: Optional[str] = None,
        stage_id: Optional[Union[str, UUID4]] = None,
        stage_name: Optional[str] = None,
    ) -> StageDB:
        return self._set_pause_state(
            project_id=project_id, project_name=project_name, stage_id=stage_id, stage_name=stage_name, pause_flag=True
        )

    def resume(
        self,
        project_id: Optional[Union[str, UUID4]] = None,
        project_name: Optional[str] = None,
        stage_id: Optional[Union[str, UUID4]] = None,
        stage_name: Optional[str] = None,
    ) -> StageDB:
        return self._set_pause_state(
            project_id=project_id, project_name=project_name, stage_id=stage_id, stage_name=stage_name, pause_flag=False
        )


def create_stage(
    project_id: Union[str, UUID4],
    name: Optional[str] = None,
    stage_type: StageType = StageType.local,
    pause: bool = False,
    rulesets: Optional[list[Rule]] = None,
    description: Optional[str] = None,
) -> Optional[StageDB]:
    """Creates a new stage.

    Args:
        project_id: The ID of the project.
        name: Name of the stage. Defaults to a generated name.
        stage_type: Type of the stage.
        pause: Whether the stage should be created in a paused state.
        rulesets: List of rules for the stage.
        description: Description for the stage.

    Returns:
        The newly created Stage.
    """
    return Stages().create(
        project_id=project_id, name=name, stage_type=stage_type, pause=pause, rulesets=rulesets, description=description
    )


def get_stage(
    project_id: Optional[Union[str, UUID4]] = None,
    project_name: Optional[str] = None,
    stage_id: Optional[Union[str, UUID4]] = None,
    stage_name: Optional[str] = None,
) -> Optional[StageDB]:
    """Retrieves a stage by its ID or name, within a given project.

    Args:
        project_id: ID of the project.
        project_name: Name of the project. If `project_id` is not provided,
                      this will be used to look up the project.
        stage_id: ID of the stage to retrieve.
        stage_name: Name of the stage to retrieve. If `stage_id` is not provided,
                    this will be used (in conjunction with project ID/name).

    Returns:
        The fetched Stage.
    """
    return Stages().get(project_id=project_id, project_name=project_name, stage_id=stage_id, stage_name=stage_name)


def update_stage(
    project_id: Optional[Union[str, UUID4]] = None,
    project_name: Optional[str] = None,
    stage_id: Optional[Union[str, UUID4]] = None,
    stage_name: Optional[str] = None,
    prioritized_rulesets: Optional[list[Rule]] = None,
) -> Optional[StageDB]:
    """Updates a stage's rulesets, creating a new version.

    Args:
        project_id: ID of the project.
        project_name: Name of the project. If `project_id` is not provided,
                      this will be used to look up the project.
        stage_id: ID of the stage to update.
        stage_name: Name of the stage to update. If `stage_id` is not provided,
                    this will be used (in conjunction with project ID/name).
        prioritized_rulesets: New list of prioritized rulesets for the stage.
                              If `None`, effectively clears existing rulesets.

    Returns:
        The updated Stage.
    """
    return Stages().update(
        project_id=project_id,
        project_name=project_name,
        stage_id=stage_id,
        stage_name=stage_name,
        prioritized_rulesets=prioritized_rulesets,
    )


def pause_stage(
    project_id: Optional[Union[str, UUID4]] = None,
    project_name: Optional[str] = None,
    stage_id: Optional[Union[str, UUID4]] = None,
    stage_name: Optional[str] = None,
) -> Optional[StageDB]:
    """Pauses the specified stage.

    Pauses a stage using either its ID or name within the context of a project.

    Args:
        project_id: ID of the project containing the stage.
        project_name: Name of the project containing the stage.
                      (Used if `project_id` is not provided).
        stage_id: ID of the stage to pause.
        stage_name: Name of the stage to pause.
                    (Used if `stage_id` is not provided).

    Returns:
        The Stage with its updated pause state.
    """
    return Stages().pause(project_id=project_id, project_name=project_name, stage_id=stage_id, stage_name=stage_name)


def resume_stage(
    project_id: Optional[Union[str, UUID4]] = None,
    project_name: Optional[str] = None,
    stage_id: Optional[Union[str, UUID4]] = None,
    stage_name: Optional[str] = None,
) -> Optional[StageDB]:
    """Resumes a previously paused stage.

    Resumes a stage using either its ID or name within the context of a project.

    Args:
        project_id: ID of the project containing the stage.
        project_name: Name of the project containing the stage.
                      (Used if `project_id` is not provided).
        stage_id: ID of the stage to resume.
        stage_name: Name of the stage to resume.
                    (Used if `stage_id` is not provided).

    Returns:
        The Stage with its updated pause state.
    """
    return Stages().resume(project_id=project_id, project_name=project_name, stage_id=stage_id, stage_name=stage_name)



from collections.abc import Mapping
from typing import TYPE_CHECKING, Any, TypeVar, Union, cast

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.base_generated_scorer_db import BaseGeneratedScorerDB
    from ..models.base_registered_scorer_db import BaseRegisteredScorerDB


T = TypeVar("T", bound="BaseScorerVersionDB")


@_attrs_define
class BaseScorerVersionDB:
    """Scorer version from the scorer_versions table.

    Attributes:
        id (str):
        version (int):
        generated_scorer (Union['BaseGeneratedScorerDB', None, Unset]):
        model_name (Union[None, Unset, str]):
        num_judges (Union[None, Unset, int]):
        registered_scorer (Union['BaseRegisteredScorerDB', None, Unset]):
    """

    id: str
    version: int
    generated_scorer: Union["BaseGeneratedScorerDB", None, Unset] = UNSET
    model_name: Union[None, Unset, str] = UNSET
    num_judges: Union[None, Unset, int] = UNSET
    registered_scorer: Union["BaseRegisteredScorerDB", None, Unset] = UNSET
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        from ..models.base_generated_scorer_db import BaseGeneratedScorerDB
        from ..models.base_registered_scorer_db import BaseRegisteredScorerDB

        id = self.id

        version = self.version

        generated_scorer: Union[None, Unset, dict[str, Any]]
        if isinstance(self.generated_scorer, Unset):
            generated_scorer = UNSET
        elif isinstance(self.generated_scorer, BaseGeneratedScorerDB):
            generated_scorer = self.generated_scorer.to_dict()
        else:
            generated_scorer = self.generated_scorer

        model_name: Union[None, Unset, str]
        if isinstance(self.model_name, Unset):
            model_name = UNSET
        else:
            model_name = self.model_name

        num_judges: Union[None, Unset, int]
        if isinstance(self.num_judges, Unset):
            num_judges = UNSET
        else:
            num_judges = self.num_judges

        registered_scorer: Union[None, Unset, dict[str, Any]]
        if isinstance(self.registered_scorer, Unset):
            registered_scorer = UNSET
        elif isinstance(self.registered_scorer, BaseRegisteredScorerDB):
            registered_scorer = self.registered_scorer.to_dict()
        else:
            registered_scorer = self.registered_scorer

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({"id": id, "version": version})
        if generated_scorer is not UNSET:
            field_dict["generated_scorer"] = generated_scorer
        if model_name is not UNSET:
            field_dict["model_name"] = model_name
        if num_judges is not UNSET:
            field_dict["num_judges"] = num_judges
        if registered_scorer is not UNSET:
            field_dict["registered_scorer"] = registered_scorer

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        from ..models.base_generated_scorer_db import BaseGeneratedScorerDB
        from ..models.base_registered_scorer_db import BaseRegisteredScorerDB

        d = dict(src_dict)
        id = d.pop("id")

        version = d.pop("version")

        def _parse_generated_scorer(data: object) -> Union["BaseGeneratedScorerDB", None, Unset]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            try:
                if not isinstance(data, dict):
                    raise TypeError()
                generated_scorer_type_0 = BaseGeneratedScorerDB.from_dict(data)

                return generated_scorer_type_0
            except:  # noqa: E722
                pass
            return cast(Union["BaseGeneratedScorerDB", None, Unset], data)

        generated_scorer = _parse_generated_scorer(d.pop("generated_scorer", UNSET))

        def _parse_model_name(data: object) -> Union[None, Unset, str]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, str], data)

        model_name = _parse_model_name(d.pop("model_name", UNSET))

        def _parse_num_judges(data: object) -> Union[None, Unset, int]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, int], data)

        num_judges = _parse_num_judges(d.pop("num_judges", UNSET))

        def _parse_registered_scorer(data: object) -> Union["BaseRegisteredScorerDB", None, Unset]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            try:
                if not isinstance(data, dict):
                    raise TypeError()
                registered_scorer_type_0 = BaseRegisteredScorerDB.from_dict(data)

                return registered_scorer_type_0
            except:  # noqa: E722
                pass
            return cast(Union["BaseRegisteredScorerDB", None, Unset], data)

        registered_scorer = _parse_registered_scorer(d.pop("registered_scorer", UNSET))

        base_scorer_version_db = cls(
            id=id,
            version=version,
            generated_scorer=generated_scorer,
            model_name=model_name,
            num_judges=num_judges,
            registered_scorer=registered_scorer,
        )

        base_scorer_version_db.additional_properties = d
        return base_scorer_version_db

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from collections.abc import Mapping
from typing import TYPE_CHECKING, Any, TypeVar, Union, cast

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..models.logging_method import LoggingMethod
from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.agent_span import AgentSpan
    from ..models.llm_span import LlmSpan
    from ..models.retriever_span import RetrieverSpan
    from ..models.tool_span import ToolSpan
    from ..models.workflow_span import WorkflowSpan


T = TypeVar("T", bound="LogSpansIngestRequest")


@_attrs_define
class LogSpansIngestRequest:
    """Request model for ingesting spans.

    Attributes:
        parent_id (str): Parent trace or span id.
        spans (list[Union['AgentSpan', 'LlmSpan', 'RetrieverSpan', 'ToolSpan', 'WorkflowSpan']]): List of spans to log.
        trace_id (str): Trace id associated with the spans.
        client_version (Union[None, Unset, str]):
        experiment_id (Union[None, Unset, str]): Experiment id associated with the traces.
        log_stream_id (Union[None, Unset, str]): Log stream id associated with the traces.
        logging_method (Union[Unset, LoggingMethod]):
        reliable (Union[Unset, bool]): Whether or not to use reliable logging.  If set to False, the method will respond
            immediately before verifying that the traces have been successfully ingested, and no error message will be
            returned if ingestion fails.  If set to True, the method will wait for the traces to be successfully ingested or
            return an error message if there is an ingestion failure. Default: False.
    """

    parent_id: str
    spans: list[Union["AgentSpan", "LlmSpan", "RetrieverSpan", "ToolSpan", "WorkflowSpan"]]
    trace_id: str
    client_version: Union[None, Unset, str] = UNSET
    experiment_id: Union[None, Unset, str] = UNSET
    log_stream_id: Union[None, Unset, str] = UNSET
    logging_method: Union[Unset, LoggingMethod] = UNSET
    reliable: Union[Unset, bool] = False
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        from ..models.agent_span import AgentSpan
        from ..models.llm_span import LlmSpan
        from ..models.retriever_span import RetrieverSpan
        from ..models.workflow_span import WorkflowSpan

        parent_id = self.parent_id

        spans = []
        for spans_item_data in self.spans:
            spans_item: dict[str, Any]
            if isinstance(spans_item_data, AgentSpan):
                spans_item = spans_item_data.to_dict()
            elif isinstance(spans_item_data, WorkflowSpan):
                spans_item = spans_item_data.to_dict()
            elif isinstance(spans_item_data, LlmSpan):
                spans_item = spans_item_data.to_dict()
            elif isinstance(spans_item_data, RetrieverSpan):
                spans_item = spans_item_data.to_dict()
            else:
                spans_item = spans_item_data.to_dict()

            spans.append(spans_item)

        trace_id = self.trace_id

        client_version: Union[None, Unset, str]
        if isinstance(self.client_version, Unset):
            client_version = UNSET
        else:
            client_version = self.client_version

        experiment_id: Union[None, Unset, str]
        if isinstance(self.experiment_id, Unset):
            experiment_id = UNSET
        else:
            experiment_id = self.experiment_id

        log_stream_id: Union[None, Unset, str]
        if isinstance(self.log_stream_id, Unset):
            log_stream_id = UNSET
        else:
            log_stream_id = self.log_stream_id

        logging_method: Union[Unset, str] = UNSET
        if not isinstance(self.logging_method, Unset):
            logging_method = self.logging_method.value

        reliable = self.reliable

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({"parent_id": parent_id, "spans": spans, "trace_id": trace_id})
        if client_version is not UNSET:
            field_dict["client_version"] = client_version
        if experiment_id is not UNSET:
            field_dict["experiment_id"] = experiment_id
        if log_stream_id is not UNSET:
            field_dict["log_stream_id"] = log_stream_id
        if logging_method is not UNSET:
            field_dict["logging_method"] = logging_method
        if reliable is not UNSET:
            field_dict["reliable"] = reliable

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        from ..models.agent_span import AgentSpan
        from ..models.llm_span import LlmSpan
        from ..models.retriever_span import RetrieverSpan
        from ..models.tool_span import ToolSpan
        from ..models.workflow_span import WorkflowSpan

        d = dict(src_dict)
        parent_id = d.pop("parent_id")

        spans = []
        _spans = d.pop("spans")
        for spans_item_data in _spans:

            def _parse_spans_item(
                data: object,
            ) -> Union["AgentSpan", "LlmSpan", "RetrieverSpan", "ToolSpan", "WorkflowSpan"]:
                try:
                    if not isinstance(data, dict):
                        raise TypeError()
                    spans_item_type_0 = AgentSpan.from_dict(data)

                    return spans_item_type_0
                except:  # noqa: E722
                    pass
                try:
                    if not isinstance(data, dict):
                        raise TypeError()
                    spans_item_type_1 = WorkflowSpan.from_dict(data)

                    return spans_item_type_1
                except:  # noqa: E722
                    pass
                try:
                    if not isinstance(data, dict):
                        raise TypeError()
                    spans_item_type_2 = LlmSpan.from_dict(data)

                    return spans_item_type_2
                except:  # noqa: E722
                    pass
                try:
                    if not isinstance(data, dict):
                        raise TypeError()
                    spans_item_type_3 = RetrieverSpan.from_dict(data)

                    return spans_item_type_3
                except:  # noqa: E722
                    pass
                if not isinstance(data, dict):
                    raise TypeError()
                spans_item_type_4 = ToolSpan.from_dict(data)

                return spans_item_type_4

            spans_item = _parse_spans_item(spans_item_data)

            spans.append(spans_item)

        trace_id = d.pop("trace_id")

        def _parse_client_version(data: object) -> Union[None, Unset, str]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, str], data)

        client_version = _parse_client_version(d.pop("client_version", UNSET))

        def _parse_experiment_id(data: object) -> Union[None, Unset, str]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, str], data)

        experiment_id = _parse_experiment_id(d.pop("experiment_id", UNSET))

        def _parse_log_stream_id(data: object) -> Union[None, Unset, str]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, str], data)

        log_stream_id = _parse_log_stream_id(d.pop("log_stream_id", UNSET))

        _logging_method = d.pop("logging_method", UNSET)
        logging_method: Union[Unset, LoggingMethod]
        if isinstance(_logging_method, Unset):
            logging_method = UNSET
        else:
            logging_method = LoggingMethod(_logging_method)

        reliable = d.pop("reliable", UNSET)

        log_spans_ingest_request = cls(
            parent_id=parent_id,
            spans=spans,
            trace_id=trace_id,
            client_version=client_version,
            experiment_id=experiment_id,
            log_stream_id=log_stream_id,
            logging_method=logging_method,
            reliable=reliable,
        )

        log_spans_ingest_request.additional_properties = d
        return log_spans_ingest_request

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from collections.abc import Mapping
from typing import Any, TypeVar, Union, cast

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

T = TypeVar("T", bound="MetricThreshold")


@_attrs_define
class MetricThreshold:
    """
    Attributes:
        buckets (Union[Unset, list[Union[float, int]]]): Threshold buckets for the column. If the column is a metric,
            these are the thresholds for the column.
        display_value_levels (Union[Unset, list[str]]): Ordered list of strings that raw values get transformed to for
            displaying.
        inverted (Union[Unset, bool]): Whether the column should be inverted for thresholds, i.e. if True, lower is
            better. Default: False.
    """

    buckets: Union[Unset, list[Union[float, int]]] = UNSET
    display_value_levels: Union[Unset, list[str]] = UNSET
    inverted: Union[Unset, bool] = False
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        buckets: Union[Unset, list[Union[float, int]]] = UNSET
        if not isinstance(self.buckets, Unset):
            buckets = []
            for buckets_item_data in self.buckets:
                buckets_item: Union[float, int]
                buckets_item = buckets_item_data
                buckets.append(buckets_item)

        display_value_levels: Union[Unset, list[str]] = UNSET
        if not isinstance(self.display_value_levels, Unset):
            display_value_levels = self.display_value_levels

        inverted = self.inverted

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if buckets is not UNSET:
            field_dict["buckets"] = buckets
        if display_value_levels is not UNSET:
            field_dict["display_value_levels"] = display_value_levels
        if inverted is not UNSET:
            field_dict["inverted"] = inverted

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        d = dict(src_dict)
        buckets = []
        _buckets = d.pop("buckets", UNSET)
        for buckets_item_data in _buckets or []:

            def _parse_buckets_item(data: object) -> Union[float, int]:
                return cast(Union[float, int], data)

            buckets_item = _parse_buckets_item(buckets_item_data)

            buckets.append(buckets_item)

        display_value_levels = cast(list[str], d.pop("display_value_levels", UNSET))

        inverted = d.pop("inverted", UNSET)

        metric_threshold = cls(buckets=buckets, display_value_levels=display_value_levels, inverted=inverted)

        metric_threshold.additional_properties = d
        return metric_threshold

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from http import HTTPStatus
from typing import Any, Optional, Union

import httpx

from ... import errors
from ...client import AuthenticatedClient, Client
from ...models.http_validation_error import HTTPValidationError
from ...models.list_group_collaborators_response import ListGroupCollaboratorsResponse
from ...types import UNSET, Response, Unset


def _get_kwargs(
    dataset_id: str, *, starting_token: Union[Unset, int] = 0, limit: Union[Unset, int] = 100
) -> dict[str, Any]:
    params: dict[str, Any] = {}

    params["starting_token"] = starting_token

    params["limit"] = limit

    params = {k: v for k, v in params.items() if v is not UNSET and v is not None}

    _kwargs: dict[str, Any] = {"method": "get", "url": f"/datasets/{dataset_id}/groups", "params": params}

    return _kwargs


def _parse_response(
    *, client: Union[AuthenticatedClient, Client], response: httpx.Response
) -> Optional[Union[HTTPValidationError, ListGroupCollaboratorsResponse]]:
    if response.status_code == 200:
        response_200 = ListGroupCollaboratorsResponse.from_dict(response.json())

        return response_200
    if response.status_code == 422:
        response_422 = HTTPValidationError.from_dict(response.json())

        return response_422
    if client.raise_on_unexpected_status:
        raise errors.UnexpectedStatus(response.status_code, response.content)
    else:
        return None


def _build_response(
    *, client: Union[AuthenticatedClient, Client], response: httpx.Response
) -> Response[Union[HTTPValidationError, ListGroupCollaboratorsResponse]]:
    return Response(
        status_code=HTTPStatus(response.status_code),
        content=response.content,
        headers=response.headers,
        parsed=_parse_response(client=client, response=response),
    )


def sync_detailed(
    dataset_id: str,
    *,
    client: AuthenticatedClient,
    starting_token: Union[Unset, int] = 0,
    limit: Union[Unset, int] = 100,
) -> Response[Union[HTTPValidationError, ListGroupCollaboratorsResponse]]:
    """List Group Dataset Collaborators

     List the groups with which the dataset has been shared.

    Args:
        dataset_id (str):
        starting_token (Union[Unset, int]):  Default: 0.
        limit (Union[Unset, int]):  Default: 100.

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Response[Union[HTTPValidationError, ListGroupCollaboratorsResponse]]
    """

    kwargs = _get_kwargs(dataset_id=dataset_id, starting_token=starting_token, limit=limit)

    response = client.get_httpx_client().request(**kwargs)

    return _build_response(client=client, response=response)


def sync(
    dataset_id: str,
    *,
    client: AuthenticatedClient,
    starting_token: Union[Unset, int] = 0,
    limit: Union[Unset, int] = 100,
) -> Optional[Union[HTTPValidationError, ListGroupCollaboratorsResponse]]:
    """List Group Dataset Collaborators

     List the groups with which the dataset has been shared.

    Args:
        dataset_id (str):
        starting_token (Union[Unset, int]):  Default: 0.
        limit (Union[Unset, int]):  Default: 100.

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Union[HTTPValidationError, ListGroupCollaboratorsResponse]
    """

    return sync_detailed(dataset_id=dataset_id, client=client, starting_token=starting_token, limit=limit).parsed


async def asyncio_detailed(
    dataset_id: str,
    *,
    client: AuthenticatedClient,
    starting_token: Union[Unset, int] = 0,
    limit: Union[Unset, int] = 100,
) -> Response[Union[HTTPValidationError, ListGroupCollaboratorsResponse]]:
    """List Group Dataset Collaborators

     List the groups with which the dataset has been shared.

    Args:
        dataset_id (str):
        starting_token (Union[Unset, int]):  Default: 0.
        limit (Union[Unset, int]):  Default: 100.

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Response[Union[HTTPValidationError, ListGroupCollaboratorsResponse]]
    """

    kwargs = _get_kwargs(dataset_id=dataset_id, starting_token=starting_token, limit=limit)

    response = await client.get_async_httpx_client().request(**kwargs)

    return _build_response(client=client, response=response)


async def asyncio(
    dataset_id: str,
    *,
    client: AuthenticatedClient,
    starting_token: Union[Unset, int] = 0,
    limit: Union[Unset, int] = 100,
) -> Optional[Union[HTTPValidationError, ListGroupCollaboratorsResponse]]:
    """List Group Dataset Collaborators

     List the groups with which the dataset has been shared.

    Args:
        dataset_id (str):
        starting_token (Union[Unset, int]):  Default: 0.
        limit (Union[Unset, int]):  Default: 100.

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Union[HTTPValidationError, ListGroupCollaboratorsResponse]
    """

    return (
        await asyncio_detailed(dataset_id=dataset_id, client=client, starting_token=starting_token, limit=limit)
    ).parsed



import datetime
from collections.abc import Mapping
from typing import TYPE_CHECKING, Any, TypeVar, Union, cast

from attrs import define as _attrs_define
from attrs import field as _attrs_field
from dateutil.parser import isoparse

from ..models.task_type import TaskType
from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.experiment_dataset import ExperimentDataset
    from ..models.experiment_response_aggregate_feedback import ExperimentResponseAggregateFeedback
    from ..models.experiment_response_aggregate_metrics import ExperimentResponseAggregateMetrics


T = TypeVar("T", bound="ExperimentResponse")


@_attrs_define
class ExperimentResponse:
    """
    Attributes:
        id (str): Galileo ID of the experiment
        project_id (str): Galileo ID of the project associated with this experiment
        task_type (TaskType): Valid task types for modeling.

            We store these as ints instead of strings because we will be looking this up in the database frequently.
        aggregate_feedback (Union[Unset, ExperimentResponseAggregateFeedback]): Aggregate feedback information related
            to the experiment
        aggregate_metrics (Union[Unset, ExperimentResponseAggregateMetrics]):
        created_at (Union[Unset, datetime.datetime]): Timestamp of the experiment's creation
        created_by (Union[None, Unset, str]):
        dataset (Union['ExperimentDataset', None, Unset]):
        name (Union[Unset, str]): Name of the experiment Default: ''.
        rank (Union[None, Unset, int]):
        ranking_score (Union[None, Unset, float]):
        updated_at (Union[None, Unset, datetime.datetime]): Timestamp of the trace or span's last update
        winner (Union[None, Unset, bool]):
    """

    id: str
    project_id: str
    task_type: TaskType
    aggregate_feedback: Union[Unset, "ExperimentResponseAggregateFeedback"] = UNSET
    aggregate_metrics: Union[Unset, "ExperimentResponseAggregateMetrics"] = UNSET
    created_at: Union[Unset, datetime.datetime] = UNSET
    created_by: Union[None, Unset, str] = UNSET
    dataset: Union["ExperimentDataset", None, Unset] = UNSET
    name: Union[Unset, str] = ""
    rank: Union[None, Unset, int] = UNSET
    ranking_score: Union[None, Unset, float] = UNSET
    updated_at: Union[None, Unset, datetime.datetime] = UNSET
    winner: Union[None, Unset, bool] = UNSET
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        from ..models.experiment_dataset import ExperimentDataset

        id = self.id

        project_id = self.project_id

        task_type = self.task_type.value

        aggregate_feedback: Union[Unset, dict[str, Any]] = UNSET
        if not isinstance(self.aggregate_feedback, Unset):
            aggregate_feedback = self.aggregate_feedback.to_dict()

        aggregate_metrics: Union[Unset, dict[str, Any]] = UNSET
        if not isinstance(self.aggregate_metrics, Unset):
            aggregate_metrics = self.aggregate_metrics.to_dict()

        created_at: Union[Unset, str] = UNSET
        if not isinstance(self.created_at, Unset):
            created_at = self.created_at.isoformat()

        created_by: Union[None, Unset, str]
        if isinstance(self.created_by, Unset):
            created_by = UNSET
        else:
            created_by = self.created_by

        dataset: Union[None, Unset, dict[str, Any]]
        if isinstance(self.dataset, Unset):
            dataset = UNSET
        elif isinstance(self.dataset, ExperimentDataset):
            dataset = self.dataset.to_dict()
        else:
            dataset = self.dataset

        name = self.name

        rank: Union[None, Unset, int]
        if isinstance(self.rank, Unset):
            rank = UNSET
        else:
            rank = self.rank

        ranking_score: Union[None, Unset, float]
        if isinstance(self.ranking_score, Unset):
            ranking_score = UNSET
        else:
            ranking_score = self.ranking_score

        updated_at: Union[None, Unset, str]
        if isinstance(self.updated_at, Unset):
            updated_at = UNSET
        elif isinstance(self.updated_at, datetime.datetime):
            updated_at = self.updated_at.isoformat()
        else:
            updated_at = self.updated_at

        winner: Union[None, Unset, bool]
        if isinstance(self.winner, Unset):
            winner = UNSET
        else:
            winner = self.winner

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({"id": id, "project_id": project_id, "task_type": task_type})
        if aggregate_feedback is not UNSET:
            field_dict["aggregate_feedback"] = aggregate_feedback
        if aggregate_metrics is not UNSET:
            field_dict["aggregate_metrics"] = aggregate_metrics
        if created_at is not UNSET:
            field_dict["created_at"] = created_at
        if created_by is not UNSET:
            field_dict["created_by"] = created_by
        if dataset is not UNSET:
            field_dict["dataset"] = dataset
        if name is not UNSET:
            field_dict["name"] = name
        if rank is not UNSET:
            field_dict["rank"] = rank
        if ranking_score is not UNSET:
            field_dict["ranking_score"] = ranking_score
        if updated_at is not UNSET:
            field_dict["updated_at"] = updated_at
        if winner is not UNSET:
            field_dict["winner"] = winner

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        from ..models.experiment_dataset import ExperimentDataset
        from ..models.experiment_response_aggregate_feedback import ExperimentResponseAggregateFeedback
        from ..models.experiment_response_aggregate_metrics import ExperimentResponseAggregateMetrics

        d = dict(src_dict)
        id = d.pop("id")

        project_id = d.pop("project_id")

        task_type = TaskType(d.pop("task_type"))

        _aggregate_feedback = d.pop("aggregate_feedback", UNSET)
        aggregate_feedback: Union[Unset, ExperimentResponseAggregateFeedback]
        if isinstance(_aggregate_feedback, Unset):
            aggregate_feedback = UNSET
        else:
            aggregate_feedback = ExperimentResponseAggregateFeedback.from_dict(_aggregate_feedback)

        _aggregate_metrics = d.pop("aggregate_metrics", UNSET)
        aggregate_metrics: Union[Unset, ExperimentResponseAggregateMetrics]
        if isinstance(_aggregate_metrics, Unset):
            aggregate_metrics = UNSET
        else:
            aggregate_metrics = ExperimentResponseAggregateMetrics.from_dict(_aggregate_metrics)

        _created_at = d.pop("created_at", UNSET)
        created_at: Union[Unset, datetime.datetime]
        if isinstance(_created_at, Unset):
            created_at = UNSET
        else:
            created_at = isoparse(_created_at)

        def _parse_created_by(data: object) -> Union[None, Unset, str]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, str], data)

        created_by = _parse_created_by(d.pop("created_by", UNSET))

        def _parse_dataset(data: object) -> Union["ExperimentDataset", None, Unset]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            try:
                if not isinstance(data, dict):
                    raise TypeError()
                dataset_type_0 = ExperimentDataset.from_dict(data)

                return dataset_type_0
            except:  # noqa: E722
                pass
            return cast(Union["ExperimentDataset", None, Unset], data)

        dataset = _parse_dataset(d.pop("dataset", UNSET))

        name = d.pop("name", UNSET)

        def _parse_rank(data: object) -> Union[None, Unset, int]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, int], data)

        rank = _parse_rank(d.pop("rank", UNSET))

        def _parse_ranking_score(data: object) -> Union[None, Unset, float]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, float], data)

        ranking_score = _parse_ranking_score(d.pop("ranking_score", UNSET))

        def _parse_updated_at(data: object) -> Union[None, Unset, datetime.datetime]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            try:
                if not isinstance(data, str):
                    raise TypeError()
                updated_at_type_0 = isoparse(data)

                return updated_at_type_0
            except:  # noqa: E722
                pass
            return cast(Union[None, Unset, datetime.datetime], data)

        updated_at = _parse_updated_at(d.pop("updated_at", UNSET))

        def _parse_winner(data: object) -> Union[None, Unset, bool]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, bool], data)

        winner = _parse_winner(d.pop("winner", UNSET))

        experiment_response = cls(
            id=id,
            project_id=project_id,
            task_type=task_type,
            aggregate_feedback=aggregate_feedback,
            aggregate_metrics=aggregate_metrics,
            created_at=created_at,
            created_by=created_by,
            dataset=dataset,
            name=name,
            rank=rank,
            ranking_score=ranking_score,
            updated_at=updated_at,
            winner=winner,
        )

        experiment_response.additional_properties = d
        return experiment_response

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



import datetime
from collections.abc import Mapping
from typing import TYPE_CHECKING, Any, Literal, TypeVar, Union, cast
from uuid import UUID

from attrs import define as _attrs_define
from attrs import field as _attrs_field
from dateutil.parser import isoparse

from ..models.scorer_type import ScorerType
from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.document import Document
    from ..models.feedback_aggregate import FeedbackAggregate
    from ..models.feedback_rating_db import FeedbackRatingDB
    from ..models.hallucination_segment import HallucinationSegment
    from ..models.metric_critique_columnar import MetricCritiqueColumnar
    from ..models.segment import Segment


T = TypeVar("T", bound="MetricSuccess")


@_attrs_define
class MetricSuccess:
    """
    Attributes:
        value (Union['Document', 'FeedbackAggregate', 'FeedbackRatingDB', 'HallucinationSegment', 'Segment', None, UUID,
            bool, datetime.datetime, float, int, list[Union['Document', 'FeedbackAggregate', 'FeedbackRatingDB',
            'HallucinationSegment', 'Segment', None, UUID, bool, datetime.datetime, float, int, str]],
            list[list[Union['Document', 'FeedbackAggregate', 'FeedbackRatingDB', 'HallucinationSegment', 'Segment', None,
            UUID, bool, datetime.datetime, float, int, str]]], list[list[list[Union['Document', 'FeedbackAggregate',
            'FeedbackRatingDB', 'HallucinationSegment', 'Segment', None, UUID, bool, datetime.datetime, float, int, str]]]],
            str]):
        cost (Union[None, Unset, float]):
        critique (Union['MetricCritiqueColumnar', None, Unset]):
        display_value (Union[None, Unset, str]):
        explanation (Union[None, Unset, str]):
        model_alias (Union[None, Unset, str]):
        num_judges (Union[None, Unset, int]):
        rationale (Union[None, Unset, str]):
        scorer_type (Union[None, ScorerType, Unset]):
        status_type (Union[Literal['success'], Unset]):  Default: 'success'.
    """

    value: Union[
        "Document",
        "FeedbackAggregate",
        "FeedbackRatingDB",
        "HallucinationSegment",
        "Segment",
        None,
        UUID,
        bool,
        datetime.datetime,
        float,
        int,
        list[
            Union[
                "Document",
                "FeedbackAggregate",
                "FeedbackRatingDB",
                "HallucinationSegment",
                "Segment",
                None,
                UUID,
                bool,
                datetime.datetime,
                float,
                int,
                str,
            ]
        ],
        list[
            list[
                Union[
                    "Document",
                    "FeedbackAggregate",
                    "FeedbackRatingDB",
                    "HallucinationSegment",
                    "Segment",
                    None,
                    UUID,
                    bool,
                    datetime.datetime,
                    float,
                    int,
                    str,
                ]
            ]
        ],
        list[
            list[
                list[
                    Union[
                        "Document",
                        "FeedbackAggregate",
                        "FeedbackRatingDB",
                        "HallucinationSegment",
                        "Segment",
                        None,
                        UUID,
                        bool,
                        datetime.datetime,
                        float,
                        int,
                        str,
                    ]
                ]
            ]
        ],
        str,
    ]
    cost: Union[None, Unset, float] = UNSET
    critique: Union["MetricCritiqueColumnar", None, Unset] = UNSET
    display_value: Union[None, Unset, str] = UNSET
    explanation: Union[None, Unset, str] = UNSET
    model_alias: Union[None, Unset, str] = UNSET
    num_judges: Union[None, Unset, int] = UNSET
    rationale: Union[None, Unset, str] = UNSET
    scorer_type: Union[None, ScorerType, Unset] = UNSET
    status_type: Union[Literal["success"], Unset] = "success"
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        from ..models.document import Document
        from ..models.feedback_aggregate import FeedbackAggregate
        from ..models.feedback_rating_db import FeedbackRatingDB
        from ..models.hallucination_segment import HallucinationSegment
        from ..models.metric_critique_columnar import MetricCritiqueColumnar
        from ..models.segment import Segment

        value: Union[
            None,
            bool,
            dict[str, Any],
            float,
            int,
            list[Union[None, bool, dict[str, Any], float, int, str]],
            list[list[Union[None, bool, dict[str, Any], float, int, str]]],
            list[list[list[Union[None, bool, dict[str, Any], float, int, str]]]],
            str,
        ]
        if isinstance(self.value, UUID):
            value = str(self.value)
        elif isinstance(self.value, datetime.datetime):
            value = self.value.isoformat()
        elif isinstance(self.value, Segment):
            value = self.value.to_dict()
        elif isinstance(self.value, HallucinationSegment):
            value = self.value.to_dict()
        elif isinstance(self.value, Document):
            value = self.value.to_dict()
        elif isinstance(self.value, FeedbackRatingDB):
            value = self.value.to_dict()
        elif isinstance(self.value, FeedbackAggregate):
            value = self.value.to_dict()
        elif isinstance(self.value, list):
            value = []
            for value_type_11_item_data in self.value:
                value_type_11_item: Union[None, bool, dict[str, Any], float, int, str]
                if isinstance(value_type_11_item_data, UUID):
                    value_type_11_item = str(value_type_11_item_data)
                elif isinstance(value_type_11_item_data, datetime.datetime):
                    value_type_11_item = value_type_11_item_data.isoformat()
                elif isinstance(value_type_11_item_data, Segment):
                    value_type_11_item = value_type_11_item_data.to_dict()
                elif isinstance(value_type_11_item_data, HallucinationSegment):
                    value_type_11_item = value_type_11_item_data.to_dict()
                elif isinstance(value_type_11_item_data, Document):
                    value_type_11_item = value_type_11_item_data.to_dict()
                elif isinstance(value_type_11_item_data, FeedbackRatingDB):
                    value_type_11_item = value_type_11_item_data.to_dict()
                elif isinstance(value_type_11_item_data, FeedbackAggregate):
                    value_type_11_item = value_type_11_item_data.to_dict()
                else:
                    value_type_11_item = value_type_11_item_data
                value.append(value_type_11_item)

        elif isinstance(self.value, list):
            value = []
            for value_type_12_item_data in self.value:
                value_type_12_item = []
                for value_type_12_item_item_data in value_type_12_item_data:
                    value_type_12_item_item: Union[None, bool, dict[str, Any], float, int, str]
                    if isinstance(value_type_12_item_item_data, UUID):
                        value_type_12_item_item = str(value_type_12_item_item_data)
                    elif isinstance(value_type_12_item_item_data, datetime.datetime):
                        value_type_12_item_item = value_type_12_item_item_data.isoformat()
                    elif isinstance(value_type_12_item_item_data, Segment):
                        value_type_12_item_item = value_type_12_item_item_data.to_dict()
                    elif isinstance(value_type_12_item_item_data, HallucinationSegment):
                        value_type_12_item_item = value_type_12_item_item_data.to_dict()
                    elif isinstance(value_type_12_item_item_data, Document):
                        value_type_12_item_item = value_type_12_item_item_data.to_dict()
                    elif isinstance(value_type_12_item_item_data, FeedbackRatingDB):
                        value_type_12_item_item = value_type_12_item_item_data.to_dict()
                    elif isinstance(value_type_12_item_item_data, FeedbackAggregate):
                        value_type_12_item_item = value_type_12_item_item_data.to_dict()
                    else:
                        value_type_12_item_item = value_type_12_item_item_data
                    value_type_12_item.append(value_type_12_item_item)

                value.append(value_type_12_item)

        elif isinstance(self.value, list):
            value = []
            for value_type_13_item_data in self.value:
                value_type_13_item = []
                for value_type_13_item_item_data in value_type_13_item_data:
                    value_type_13_item_item = []
                    for value_type_13_item_item_item_data in value_type_13_item_item_data:
                        value_type_13_item_item_item: Union[None, bool, dict[str, Any], float, int, str]
                        if isinstance(value_type_13_item_item_item_data, UUID):
                            value_type_13_item_item_item = str(value_type_13_item_item_item_data)
                        elif isinstance(value_type_13_item_item_item_data, datetime.datetime):
                            value_type_13_item_item_item = value_type_13_item_item_item_data.isoformat()
                        elif isinstance(value_type_13_item_item_item_data, Segment):
                            value_type_13_item_item_item = value_type_13_item_item_item_data.to_dict()
                        elif isinstance(value_type_13_item_item_item_data, HallucinationSegment):
                            value_type_13_item_item_item = value_type_13_item_item_item_data.to_dict()
                        elif isinstance(value_type_13_item_item_item_data, Document):
                            value_type_13_item_item_item = value_type_13_item_item_item_data.to_dict()
                        elif isinstance(value_type_13_item_item_item_data, FeedbackRatingDB):
                            value_type_13_item_item_item = value_type_13_item_item_item_data.to_dict()
                        elif isinstance(value_type_13_item_item_item_data, FeedbackAggregate):
                            value_type_13_item_item_item = value_type_13_item_item_item_data.to_dict()
                        else:
                            value_type_13_item_item_item = value_type_13_item_item_item_data
                        value_type_13_item_item.append(value_type_13_item_item_item)

                    value_type_13_item.append(value_type_13_item_item)

                value.append(value_type_13_item)

        else:
            value = self.value

        cost: Union[None, Unset, float]
        if isinstance(self.cost, Unset):
            cost = UNSET
        else:
            cost = self.cost

        critique: Union[None, Unset, dict[str, Any]]
        if isinstance(self.critique, Unset):
            critique = UNSET
        elif isinstance(self.critique, MetricCritiqueColumnar):
            critique = self.critique.to_dict()
        else:
            critique = self.critique

        display_value: Union[None, Unset, str]
        if isinstance(self.display_value, Unset):
            display_value = UNSET
        else:
            display_value = self.display_value

        explanation: Union[None, Unset, str]
        if isinstance(self.explanation, Unset):
            explanation = UNSET
        else:
            explanation = self.explanation

        model_alias: Union[None, Unset, str]
        if isinstance(self.model_alias, Unset):
            model_alias = UNSET
        else:
            model_alias = self.model_alias

        num_judges: Union[None, Unset, int]
        if isinstance(self.num_judges, Unset):
            num_judges = UNSET
        else:
            num_judges = self.num_judges

        rationale: Union[None, Unset, str]
        if isinstance(self.rationale, Unset):
            rationale = UNSET
        else:
            rationale = self.rationale

        scorer_type: Union[None, Unset, str]
        if isinstance(self.scorer_type, Unset):
            scorer_type = UNSET
        elif isinstance(self.scorer_type, ScorerType):
            scorer_type = self.scorer_type.value
        else:
            scorer_type = self.scorer_type

        status_type = self.status_type

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({"value": value})
        if cost is not UNSET:
            field_dict["cost"] = cost
        if critique is not UNSET:
            field_dict["critique"] = critique
        if display_value is not UNSET:
            field_dict["display_value"] = display_value
        if explanation is not UNSET:
            field_dict["explanation"] = explanation
        if model_alias is not UNSET:
            field_dict["model_alias"] = model_alias
        if num_judges is not UNSET:
            field_dict["num_judges"] = num_judges
        if rationale is not UNSET:
            field_dict["rationale"] = rationale
        if scorer_type is not UNSET:
            field_dict["scorer_type"] = scorer_type
        if status_type is not UNSET:
            field_dict["status_type"] = status_type

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        from ..models.document import Document
        from ..models.feedback_aggregate import FeedbackAggregate
        from ..models.feedback_rating_db import FeedbackRatingDB
        from ..models.hallucination_segment import HallucinationSegment
        from ..models.metric_critique_columnar import MetricCritiqueColumnar
        from ..models.segment import Segment

        d = dict(src_dict)

        def _parse_value(
            data: object,
        ) -> Union[
            "Document",
            "FeedbackAggregate",
            "FeedbackRatingDB",
            "HallucinationSegment",
            "Segment",
            None,
            UUID,
            bool,
            datetime.datetime,
            float,
            int,
            list[
                Union[
                    "Document",
                    "FeedbackAggregate",
                    "FeedbackRatingDB",
                    "HallucinationSegment",
                    "Segment",
                    None,
                    UUID,
                    bool,
                    datetime.datetime,
                    float,
                    int,
                    str,
                ]
            ],
            list[
                list[
                    Union[
                        "Document",
                        "FeedbackAggregate",
                        "FeedbackRatingDB",
                        "HallucinationSegment",
                        "Segment",
                        None,
                        UUID,
                        bool,
                        datetime.datetime,
                        float,
                        int,
                        str,
                    ]
                ]
            ],
            list[
                list[
                    list[
                        Union[
                            "Document",
                            "FeedbackAggregate",
                            "FeedbackRatingDB",
                            "HallucinationSegment",
                            "Segment",
                            None,
                            UUID,
                            bool,
                            datetime.datetime,
                            float,
                            int,
                            str,
                        ]
                    ]
                ]
            ],
            str,
        ]:
            if data is None:
                return data
            try:
                if not isinstance(data, str):
                    raise TypeError()
                value_type_4 = UUID(data)

                return value_type_4
            except:  # noqa: E722
                pass
            try:
                if not isinstance(data, str):
                    raise TypeError()
                value_type_5 = isoparse(data)

                return value_type_5
            except:  # noqa: E722
                pass
            try:
                if not isinstance(data, dict):
                    raise TypeError()
                value_type_6 = Segment.from_dict(data)

                return value_type_6
            except:  # noqa: E722
                pass
            try:
                if not isinstance(data, dict):
                    raise TypeError()
                value_type_7 = HallucinationSegment.from_dict(data)

                return value_type_7
            except:  # noqa: E722
                pass
            try:
                if not isinstance(data, dict):
                    raise TypeError()
                value_type_8 = Document.from_dict(data)

                return value_type_8
            except:  # noqa: E722
                pass
            try:
                if not isinstance(data, dict):
                    raise TypeError()
                value_type_9 = FeedbackRatingDB.from_dict(data)

                return value_type_9
            except:  # noqa: E722
                pass
            try:
                if not isinstance(data, dict):
                    raise TypeError()
                value_type_10 = FeedbackAggregate.from_dict(data)

                return value_type_10
            except:  # noqa: E722
                pass
            try:
                if not isinstance(data, list):
                    raise TypeError()
                value_type_11 = []
                _value_type_11 = data
                for value_type_11_item_data in _value_type_11:

                    def _parse_value_type_11_item(
                        data: object,
                    ) -> Union[
                        "Document",
                        "FeedbackAggregate",
                        "FeedbackRatingDB",
                        "HallucinationSegment",
                        "Segment",
                        None,
                        UUID,
                        bool,
                        datetime.datetime,
                        float,
                        int,
                        str,
                    ]:
                        if data is None:
                            return data
                        try:
                            if not isinstance(data, str):
                                raise TypeError()
                            value_type_11_item_type_4 = UUID(data)

                            return value_type_11_item_type_4
                        except:  # noqa: E722
                            pass
                        try:
                            if not isinstance(data, str):
                                raise TypeError()
                            value_type_11_item_type_5 = isoparse(data)

                            return value_type_11_item_type_5
                        except:  # noqa: E722
                            pass
                        try:
                            if not isinstance(data, dict):
                                raise TypeError()
                            value_type_11_item_type_6 = Segment.from_dict(data)

                            return value_type_11_item_type_6
                        except:  # noqa: E722
                            pass
                        try:
                            if not isinstance(data, dict):
                                raise TypeError()
                            value_type_11_item_type_7 = HallucinationSegment.from_dict(data)

                            return value_type_11_item_type_7
                        except:  # noqa: E722
                            pass
                        try:
                            if not isinstance(data, dict):
                                raise TypeError()
                            value_type_11_item_type_8 = Document.from_dict(data)

                            return value_type_11_item_type_8
                        except:  # noqa: E722
                            pass
                        try:
                            if not isinstance(data, dict):
                                raise TypeError()
                            value_type_11_item_type_9 = FeedbackRatingDB.from_dict(data)

                            return value_type_11_item_type_9
                        except:  # noqa: E722
                            pass
                        try:
                            if not isinstance(data, dict):
                                raise TypeError()
                            value_type_11_item_type_10 = FeedbackAggregate.from_dict(data)

                            return value_type_11_item_type_10
                        except:  # noqa: E722
                            pass
                        return cast(
                            Union[
                                "Document",
                                "FeedbackAggregate",
                                "FeedbackRatingDB",
                                "HallucinationSegment",
                                "Segment",
                                None,
                                UUID,
                                bool,
                                datetime.datetime,
                                float,
                                int,
                                str,
                            ],
                            data,
                        )

                    value_type_11_item = _parse_value_type_11_item(value_type_11_item_data)

                    value_type_11.append(value_type_11_item)

                return value_type_11
            except:  # noqa: E722
                pass
            try:
                if not isinstance(data, list):
                    raise TypeError()
                value_type_12 = []
                _value_type_12 = data
                for value_type_12_item_data in _value_type_12:
                    value_type_12_item = []
                    _value_type_12_item = value_type_12_item_data
                    for value_type_12_item_item_data in _value_type_12_item:

                        def _parse_value_type_12_item_item(
                            data: object,
                        ) -> Union[
                            "Document",
                            "FeedbackAggregate",
                            "FeedbackRatingDB",
                            "HallucinationSegment",
                            "Segment",
                            None,
                            UUID,
                            bool,
                            datetime.datetime,
                            float,
                            int,
                            str,
                        ]:
                            if data is None:
                                return data
                            try:
                                if not isinstance(data, str):
                                    raise TypeError()
                                value_type_12_item_item_type_4 = UUID(data)

                                return value_type_12_item_item_type_4
                            except:  # noqa: E722
                                pass
                            try:
                                if not isinstance(data, str):
                                    raise TypeError()
                                value_type_12_item_item_type_5 = isoparse(data)

                                return value_type_12_item_item_type_5
                            except:  # noqa: E722
                                pass
                            try:
                                if not isinstance(data, dict):
                                    raise TypeError()
                                value_type_12_item_item_type_6 = Segment.from_dict(data)

                                return value_type_12_item_item_type_6
                            except:  # noqa: E722
                                pass
                            try:
                                if not isinstance(data, dict):
                                    raise TypeError()
                                value_type_12_item_item_type_7 = HallucinationSegment.from_dict(data)

                                return value_type_12_item_item_type_7
                            except:  # noqa: E722
                                pass
                            try:
                                if not isinstance(data, dict):
                                    raise TypeError()
                                value_type_12_item_item_type_8 = Document.from_dict(data)

                                return value_type_12_item_item_type_8
                            except:  # noqa: E722
                                pass
                            try:
                                if not isinstance(data, dict):
                                    raise TypeError()
                                value_type_12_item_item_type_9 = FeedbackRatingDB.from_dict(data)

                                return value_type_12_item_item_type_9
                            except:  # noqa: E722
                                pass
                            try:
                                if not isinstance(data, dict):
                                    raise TypeError()
                                value_type_12_item_item_type_10 = FeedbackAggregate.from_dict(data)

                                return value_type_12_item_item_type_10
                            except:  # noqa: E722
                                pass
                            return cast(
                                Union[
                                    "Document",
                                    "FeedbackAggregate",
                                    "FeedbackRatingDB",
                                    "HallucinationSegment",
                                    "Segment",
                                    None,
                                    UUID,
                                    bool,
                                    datetime.datetime,
                                    float,
                                    int,
                                    str,
                                ],
                                data,
                            )

                        value_type_12_item_item = _parse_value_type_12_item_item(value_type_12_item_item_data)

                        value_type_12_item.append(value_type_12_item_item)

                    value_type_12.append(value_type_12_item)

                return value_type_12
            except:  # noqa: E722
                pass
            try:
                if not isinstance(data, list):
                    raise TypeError()
                value_type_13 = []
                _value_type_13 = data
                for value_type_13_item_data in _value_type_13:
                    value_type_13_item = []
                    _value_type_13_item = value_type_13_item_data
                    for value_type_13_item_item_data in _value_type_13_item:
                        value_type_13_item_item = []
                        _value_type_13_item_item = value_type_13_item_item_data
                        for value_type_13_item_item_item_data in _value_type_13_item_item:

                            def _parse_value_type_13_item_item_item(
                                data: object,
                            ) -> Union[
                                "Document",
                                "FeedbackAggregate",
                                "FeedbackRatingDB",
                                "HallucinationSegment",
                                "Segment",
                                None,
                                UUID,
                                bool,
                                datetime.datetime,
                                float,
                                int,
                                str,
                            ]:
                                if data is None:
                                    return data
                                try:
                                    if not isinstance(data, str):
                                        raise TypeError()
                                    value_type_13_item_item_item_type_4 = UUID(data)

                                    return value_type_13_item_item_item_type_4
                                except:  # noqa: E722
                                    pass
                                try:
                                    if not isinstance(data, str):
                                        raise TypeError()
                                    value_type_13_item_item_item_type_5 = isoparse(data)

                                    return value_type_13_item_item_item_type_5
                                except:  # noqa: E722
                                    pass
                                try:
                                    if not isinstance(data, dict):
                                        raise TypeError()
                                    value_type_13_item_item_item_type_6 = Segment.from_dict(data)

                                    return value_type_13_item_item_item_type_6
                                except:  # noqa: E722
                                    pass
                                try:
                                    if not isinstance(data, dict):
                                        raise TypeError()
                                    value_type_13_item_item_item_type_7 = HallucinationSegment.from_dict(data)

                                    return value_type_13_item_item_item_type_7
                                except:  # noqa: E722
                                    pass
                                try:
                                    if not isinstance(data, dict):
                                        raise TypeError()
                                    value_type_13_item_item_item_type_8 = Document.from_dict(data)

                                    return value_type_13_item_item_item_type_8
                                except:  # noqa: E722
                                    pass
                                try:
                                    if not isinstance(data, dict):
                                        raise TypeError()
                                    value_type_13_item_item_item_type_9 = FeedbackRatingDB.from_dict(data)

                                    return value_type_13_item_item_item_type_9
                                except:  # noqa: E722
                                    pass
                                try:
                                    if not isinstance(data, dict):
                                        raise TypeError()
                                    value_type_13_item_item_item_type_10 = FeedbackAggregate.from_dict(data)

                                    return value_type_13_item_item_item_type_10
                                except:  # noqa: E722
                                    pass
                                return cast(
                                    Union[
                                        "Document",
                                        "FeedbackAggregate",
                                        "FeedbackRatingDB",
                                        "HallucinationSegment",
                                        "Segment",
                                        None,
                                        UUID,
                                        bool,
                                        datetime.datetime,
                                        float,
                                        int,
                                        str,
                                    ],
                                    data,
                                )

                            value_type_13_item_item_item = _parse_value_type_13_item_item_item(
                                value_type_13_item_item_item_data
                            )

                            value_type_13_item_item.append(value_type_13_item_item_item)

                        value_type_13_item.append(value_type_13_item_item)

                    value_type_13.append(value_type_13_item)

                return value_type_13
            except:  # noqa: E722
                pass
            return cast(
                Union[
                    "Document",
                    "FeedbackAggregate",
                    "FeedbackRatingDB",
                    "HallucinationSegment",
                    "Segment",
                    None,
                    UUID,
                    bool,
                    datetime.datetime,
                    float,
                    int,
                    list[
                        Union[
                            "Document",
                            "FeedbackAggregate",
                            "FeedbackRatingDB",
                            "HallucinationSegment",
                            "Segment",
                            None,
                            UUID,
                            bool,
                            datetime.datetime,
                            float,
                            int,
                            str,
                        ]
                    ],
                    list[
                        list[
                            Union[
                                "Document",
                                "FeedbackAggregate",
                                "FeedbackRatingDB",
                                "HallucinationSegment",
                                "Segment",
                                None,
                                UUID,
                                bool,
                                datetime.datetime,
                                float,
                                int,
                                str,
                            ]
                        ]
                    ],
                    list[
                        list[
                            list[
                                Union[
                                    "Document",
                                    "FeedbackAggregate",
                                    "FeedbackRatingDB",
                                    "HallucinationSegment",
                                    "Segment",
                                    None,
                                    UUID,
                                    bool,
                                    datetime.datetime,
                                    float,
                                    int,
                                    str,
                                ]
                            ]
                        ]
                    ],
                    str,
                ],
                data,
            )

        value = _parse_value(d.pop("value"))

        def _parse_cost(data: object) -> Union[None, Unset, float]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, float], data)

        cost = _parse_cost(d.pop("cost", UNSET))

        def _parse_critique(data: object) -> Union["MetricCritiqueColumnar", None, Unset]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            try:
                if not isinstance(data, dict):
                    raise TypeError()
                critique_type_0 = MetricCritiqueColumnar.from_dict(data)

                return critique_type_0
            except:  # noqa: E722
                pass
            return cast(Union["MetricCritiqueColumnar", None, Unset], data)

        critique = _parse_critique(d.pop("critique", UNSET))

        def _parse_display_value(data: object) -> Union[None, Unset, str]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, str], data)

        display_value = _parse_display_value(d.pop("display_value", UNSET))

        def _parse_explanation(data: object) -> Union[None, Unset, str]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, str], data)

        explanation = _parse_explanation(d.pop("explanation", UNSET))

        def _parse_model_alias(data: object) -> Union[None, Unset, str]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, str], data)

        model_alias = _parse_model_alias(d.pop("model_alias", UNSET))

        def _parse_num_judges(data: object) -> Union[None, Unset, int]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, int], data)

        num_judges = _parse_num_judges(d.pop("num_judges", UNSET))

        def _parse_rationale(data: object) -> Union[None, Unset, str]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, str], data)

        rationale = _parse_rationale(d.pop("rationale", UNSET))

        def _parse_scorer_type(data: object) -> Union[None, ScorerType, Unset]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            try:
                if not isinstance(data, str):
                    raise TypeError()
                scorer_type_type_0 = ScorerType(data)

                return scorer_type_type_0
            except:  # noqa: E722
                pass
            return cast(Union[None, ScorerType, Unset], data)

        scorer_type = _parse_scorer_type(d.pop("scorer_type", UNSET))

        status_type = cast(Union[Literal["success"], Unset], d.pop("status_type", UNSET))
        if status_type != "success" and not isinstance(status_type, Unset):
            raise ValueError(f"status_type must match const 'success', got '{status_type}'")

        metric_success = cls(
            value=value,
            cost=cost,
            critique=critique,
            display_value=display_value,
            explanation=explanation,
            model_alias=model_alias,
            num_judges=num_judges,
            rationale=rationale,
            scorer_type=scorer_type,
            status_type=status_type,
        )

        metric_success.additional_properties = d
        return metric_success

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



import json
from functools import cached_property
from typing import Any, Optional

from pydantic import BaseModel, field_validator


class DatasetRecord(BaseModel):
    id: Optional[str] = None
    input: str
    output: Optional[str] = None
    metadata: Optional[dict[str, str]] = None

    @field_validator("input", mode="before")
    @classmethod
    def validate_input(cls, value: Any) -> str:
        if not isinstance(value, str):
            value = json.dumps(value)
        return value

    @field_validator("output", mode="before")
    @classmethod
    def validate_output(cls, value: Any) -> Optional[str]:
        if value is None:
            return None
        if not isinstance(value, str):
            value = json.dumps(value)
        return value

    @field_validator("metadata", mode="before")
    @classmethod
    def validate_metadata(cls, value: Any) -> Optional[dict[str, str]]:
        if value is None:
            return None
        if isinstance(value, str):
            try:
                value = json.loads(value)
            except json.decoder.JSONDecodeError:
                value = {"metadata": value}
        if not isinstance(value, dict):
            raise ValueError("Dataset metadata field must be either a string or dictionary")
        for key, v in value.items():
            if not isinstance(v, str):
                raise ValueError("Dataset metadata field dictionary values must be strings")
        return value

    @cached_property
    def deserialized_input(self) -> Any:
        try:
            return json.loads(self.input)
        except json.decoder.JSONDecodeError:
            return self.input

    @cached_property
    def deserialized_output(self) -> Optional[Any]:
        if self.output is None:
            return None
        try:
            return json.loads(self.output)
        except json.decoder.JSONDecodeError:
            return self.output



from collections.abc import Mapping
from typing import Any, Literal, TypeVar, Union, cast

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..models.project_runs_filter_operator import ProjectRunsFilterOperator
from ..types import UNSET, Unset

T = TypeVar("T", bound="ProjectRunsFilter")


@_attrs_define
class ProjectRunsFilter:
    """
    Attributes:
        operator (ProjectRunsFilterOperator):
        value (Union[float, int, list[float], list[int]]):
        name (Union[Literal['runs'], Unset]):  Default: 'runs'.
    """

    operator: ProjectRunsFilterOperator
    value: Union[float, int, list[float], list[int]]
    name: Union[Literal["runs"], Unset] = "runs"
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        operator = self.operator.value

        value: Union[float, int, list[float], list[int]]
        if isinstance(self.value, list):
            value = self.value

        elif isinstance(self.value, list):
            value = self.value

        else:
            value = self.value

        name = self.name

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({"operator": operator, "value": value})
        if name is not UNSET:
            field_dict["name"] = name

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        d = dict(src_dict)
        operator = ProjectRunsFilterOperator(d.pop("operator"))

        def _parse_value(data: object) -> Union[float, int, list[float], list[int]]:
            try:
                if not isinstance(data, list):
                    raise TypeError()
                value_type_2 = cast(list[int], data)

                return value_type_2
            except:  # noqa: E722
                pass
            try:
                if not isinstance(data, list):
                    raise TypeError()
                value_type_3 = cast(list[float], data)

                return value_type_3
            except:  # noqa: E722
                pass
            return cast(Union[float, int, list[float], list[int]], data)

        value = _parse_value(d.pop("value"))

        name = cast(Union[Literal["runs"], Unset], d.pop("name", UNSET))
        if name != "runs" and not isinstance(name, Unset):
            raise ValueError(f"name must match const 'runs', got '{name}'")

        project_runs_filter = cls(operator=operator, value=value, name=name)

        project_runs_filter.additional_properties = d
        return project_runs_filter

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from collections.abc import Mapping
from typing import TYPE_CHECKING, Any, Literal, TypeVar, Union, cast

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.star_aggregate_counts import StarAggregateCounts


T = TypeVar("T", bound="StarAggregate")


@_attrs_define
class StarAggregate:
    """
    Attributes:
        average (float):
        counts (StarAggregateCounts):
        unrated_count (int):
        feedback_type (Union[Literal['star'], Unset]):  Default: 'star'.
    """

    average: float
    counts: "StarAggregateCounts"
    unrated_count: int
    feedback_type: Union[Literal["star"], Unset] = "star"
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        average = self.average

        counts = self.counts.to_dict()

        unrated_count = self.unrated_count

        feedback_type = self.feedback_type

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({"average": average, "counts": counts, "unrated_count": unrated_count})
        if feedback_type is not UNSET:
            field_dict["feedback_type"] = feedback_type

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        from ..models.star_aggregate_counts import StarAggregateCounts

        d = dict(src_dict)
        average = d.pop("average")

        counts = StarAggregateCounts.from_dict(d.pop("counts"))

        unrated_count = d.pop("unrated_count")

        feedback_type = cast(Union[Literal["star"], Unset], d.pop("feedback_type", UNSET))
        if feedback_type != "star" and not isinstance(feedback_type, Unset):
            raise ValueError(f"feedback_type must match const 'star', got '{feedback_type}'")

        star_aggregate = cls(average=average, counts=counts, unrated_count=unrated_count, feedback_type=feedback_type)

        star_aggregate.additional_properties = d
        return star_aggregate

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from collections.abc import Mapping
from typing import TYPE_CHECKING, Any, TypeVar, Union, cast

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.dataset_content_filter import DatasetContentFilter
    from ..models.dataset_content_sort_clause import DatasetContentSortClause


T = TypeVar("T", bound="QueryDatasetParams")


@_attrs_define
class QueryDatasetParams:
    """
    Attributes:
        filters (Union[Unset, list['DatasetContentFilter']]):
        sort (Union['DatasetContentSortClause', None, Unset]):
    """

    filters: Union[Unset, list["DatasetContentFilter"]] = UNSET
    sort: Union["DatasetContentSortClause", None, Unset] = UNSET
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        from ..models.dataset_content_sort_clause import DatasetContentSortClause

        filters: Union[Unset, list[dict[str, Any]]] = UNSET
        if not isinstance(self.filters, Unset):
            filters = []
            for filters_item_data in self.filters:
                filters_item = filters_item_data.to_dict()
                filters.append(filters_item)

        sort: Union[None, Unset, dict[str, Any]]
        if isinstance(self.sort, Unset):
            sort = UNSET
        elif isinstance(self.sort, DatasetContentSortClause):
            sort = self.sort.to_dict()
        else:
            sort = self.sort

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if filters is not UNSET:
            field_dict["filters"] = filters
        if sort is not UNSET:
            field_dict["sort"] = sort

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        from ..models.dataset_content_filter import DatasetContentFilter
        from ..models.dataset_content_sort_clause import DatasetContentSortClause

        d = dict(src_dict)
        filters = []
        _filters = d.pop("filters", UNSET)
        for filters_item_data in _filters or []:
            filters_item = DatasetContentFilter.from_dict(filters_item_data)

            filters.append(filters_item)

        def _parse_sort(data: object) -> Union["DatasetContentSortClause", None, Unset]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            try:
                if not isinstance(data, dict):
                    raise TypeError()
                sort_type_0 = DatasetContentSortClause.from_dict(data)

                return sort_type_0
            except:  # noqa: E722
                pass
            return cast(Union["DatasetContentSortClause", None, Unset], data)

        sort = _parse_sort(d.pop("sort", UNSET))

        query_dataset_params = cls(filters=filters, sort=sort)

        query_dataset_params.additional_properties = d
        return query_dataset_params

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from collections.abc import Sequence
from typing import Optional

from langchain_core.runnables.base import Runnable
from langchain_core.tools import BaseTool
from pydantic import UUID4, BaseModel, ConfigDict, Field
from pydantic.v1 import BaseModel as BaseModelV1

from galileo.constants.protect import TIMEOUT_SECS
from galileo.protect import ainvoke, invoke
from galileo_core.schemas.protect.execution_status import ExecutionStatus
from galileo_core.schemas.protect.payload import Payload as CorePayload
from galileo_core.schemas.protect.response import Response
from galileo_core.schemas.protect.ruleset import Ruleset


class ProtectToolInputSchema(BaseModelV1):
    input: Optional[str] = None
    output: Optional[str] = None


class ProtectTool(BaseTool):
    name: str = "GalileoProtect"
    description: str = (
        "Protect your LLM applications from harmful content using Galileo Protect. "
        "This tool is a wrapper around Galileo's Protect API, can be used to scan text "
        "for harmful content, and can be used to trigger actions based on the results."
        "The tool can be used on the input text or output text, and can be configured "
        "with a set of rulesets to evaluate on."
    )
    # langchain requires pydantic v1
    args_schema: type[BaseModelV1] = ProtectToolInputSchema

    prioritized_rulesets: Optional[Sequence[Ruleset]] = None
    project_id: Optional[UUID4] = None
    project_name: Optional[str] = None
    stage_name: Optional[str] = None
    stage_id: Optional[UUID4] = None
    stage_version: Optional[int] = None
    timeout: float = TIMEOUT_SECS

    def _run(self, input: Optional[str] = None, output: Optional[str] = None) -> str:
        """
        Apply the tool synchronously.

        We serialize the response to JSON because that's what `langchain_core` expects
        for tools.
        """
        api_payload = CorePayload(input=input, output=output)
        api_response = invoke(
            payload=api_payload,
            prioritized_rulesets=self.prioritized_rulesets,
            project_name=self.project_name,
            project_id=self.project_id,
            stage_name=self.stage_name,
            stage_id=self.stage_id,
            stage_version=self.stage_version,
            timeout=self.timeout,
        )
        return api_response.model_dump_json()

    async def _arun(self, input: Optional[str] = None, output: Optional[str] = None) -> str:
        """
        Apply the tool asynchronously.

        We serialize the response to JSON because that's what `langchain_core` expects
        for tools.
        """
        api_payload = CorePayload(input=input, output=output)
        api_response = await ainvoke(
            payload=api_payload,
            prioritized_rulesets=self.prioritized_rulesets,
            project_name=self.project_name,
            project_id=self.project_id,
            stage_name=self.stage_name,
            stage_id=self.stage_id,
            stage_version=self.stage_version,
            timeout=self.timeout,
        )
        return api_response.model_dump_json()


class ProtectParser(BaseModel):
    chain: Runnable = Field(..., description="The chain to trigger if the Protect invocation is not triggered.")
    ignore_trigger: bool = Field(
        default=False,
        description="Ignore the status of the Protect invocation and always trigger the rest of the chain.",
    )
    echo_output: bool = Field(default=False, description="Echo the output of the Protect invocation.")

    model_config = ConfigDict(arbitrary_types_allowed=True)

    def parser(self, response_raw_json: str) -> str:
        try:
            response = Response.model_validate_json(response_raw_json)
        except Exception:
            return self.chain.invoke(response_raw_json)
        text = response.text

        if self.echo_output:
            print(f"> Raw response: {text}")
        if response.status == ExecutionStatus.triggered and not self.ignore_trigger:
            return text
        else:
            return self.chain.invoke(text)



from collections.abc import Mapping
from typing import TYPE_CHECKING, Any, TypeVar, Union

from attrs import define as _attrs_define
from attrs import field as _attrs_field

if TYPE_CHECKING:
    from ..models.like_dislike_aggregate import LikeDislikeAggregate
    from ..models.score_aggregate import ScoreAggregate
    from ..models.star_aggregate import StarAggregate
    from ..models.tags_aggregate import TagsAggregate


T = TypeVar("T", bound="FeedbackAggregate")


@_attrs_define
class FeedbackAggregate:
    """
    Attributes:
        aggregate (Union['LikeDislikeAggregate', 'ScoreAggregate', 'StarAggregate', 'TagsAggregate']):
    """

    aggregate: Union["LikeDislikeAggregate", "ScoreAggregate", "StarAggregate", "TagsAggregate"]
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        from ..models.like_dislike_aggregate import LikeDislikeAggregate
        from ..models.score_aggregate import ScoreAggregate
        from ..models.star_aggregate import StarAggregate

        aggregate: dict[str, Any]
        if isinstance(self.aggregate, LikeDislikeAggregate):
            aggregate = self.aggregate.to_dict()
        elif isinstance(self.aggregate, StarAggregate):
            aggregate = self.aggregate.to_dict()
        elif isinstance(self.aggregate, ScoreAggregate):
            aggregate = self.aggregate.to_dict()
        else:
            aggregate = self.aggregate.to_dict()

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({"aggregate": aggregate})

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        from ..models.like_dislike_aggregate import LikeDislikeAggregate
        from ..models.score_aggregate import ScoreAggregate
        from ..models.star_aggregate import StarAggregate
        from ..models.tags_aggregate import TagsAggregate

        d = dict(src_dict)

        def _parse_aggregate(
            data: object,
        ) -> Union["LikeDislikeAggregate", "ScoreAggregate", "StarAggregate", "TagsAggregate"]:
            try:
                if not isinstance(data, dict):
                    raise TypeError()
                aggregate_type_0 = LikeDislikeAggregate.from_dict(data)

                return aggregate_type_0
            except:  # noqa: E722
                pass
            try:
                if not isinstance(data, dict):
                    raise TypeError()
                aggregate_type_1 = StarAggregate.from_dict(data)

                return aggregate_type_1
            except:  # noqa: E722
                pass
            try:
                if not isinstance(data, dict):
                    raise TypeError()
                aggregate_type_2 = ScoreAggregate.from_dict(data)

                return aggregate_type_2
            except:  # noqa: E722
                pass
            if not isinstance(data, dict):
                raise TypeError()
            aggregate_type_3 = TagsAggregate.from_dict(data)

            return aggregate_type_3

        aggregate = _parse_aggregate(d.pop("aggregate"))

        feedback_aggregate = cls(aggregate=aggregate)

        feedback_aggregate.additional_properties = d
        return feedback_aggregate

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from collections.abc import Mapping
from typing import TYPE_CHECKING, Any, TypeVar, Union, cast

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.column_mapping import ColumnMapping


T = TypeVar("T", bound="PreviewDatasetRequest")


@_attrs_define
class PreviewDatasetRequest:
    """
    Attributes:
        column_mapping (Union['ColumnMapping', None, Unset]):
    """

    column_mapping: Union["ColumnMapping", None, Unset] = UNSET
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        from ..models.column_mapping import ColumnMapping

        column_mapping: Union[None, Unset, dict[str, Any]]
        if isinstance(self.column_mapping, Unset):
            column_mapping = UNSET
        elif isinstance(self.column_mapping, ColumnMapping):
            column_mapping = self.column_mapping.to_dict()
        else:
            column_mapping = self.column_mapping

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if column_mapping is not UNSET:
            field_dict["column_mapping"] = column_mapping

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        from ..models.column_mapping import ColumnMapping

        d = dict(src_dict)

        def _parse_column_mapping(data: object) -> Union["ColumnMapping", None, Unset]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            try:
                if not isinstance(data, dict):
                    raise TypeError()
                column_mapping_type_0 = ColumnMapping.from_dict(data)

                return column_mapping_type_0
            except:  # noqa: E722
                pass
            return cast(Union["ColumnMapping", None, Unset], data)

        column_mapping = _parse_column_mapping(d.pop("column_mapping", UNSET))

        preview_dataset_request = cls(column_mapping=column_mapping)

        preview_dataset_request.additional_properties = d
        return preview_dataset_request

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from collections.abc import Mapping
from typing import TYPE_CHECKING, Any, Literal, TypeVar, Union, cast

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

if TYPE_CHECKING:
    from ..models.dataset_append_row_values import DatasetAppendRowValues


T = TypeVar("T", bound="DatasetAppendRow")


@_attrs_define
class DatasetAppendRow:
    """
    Attributes:
        values (DatasetAppendRowValues):
        edit_type (Union[Literal['append_row'], Unset]):  Default: 'append_row'.
        row_id (Union[None, Unset, str]):
    """

    values: "DatasetAppendRowValues"
    edit_type: Union[Literal["append_row"], Unset] = "append_row"
    row_id: Union[None, Unset, str] = UNSET
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        values = self.values.to_dict()

        edit_type = self.edit_type

        row_id: Union[None, Unset, str]
        if isinstance(self.row_id, Unset):
            row_id = UNSET
        else:
            row_id = self.row_id

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({"values": values})
        if edit_type is not UNSET:
            field_dict["edit_type"] = edit_type
        if row_id is not UNSET:
            field_dict["row_id"] = row_id

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        from ..models.dataset_append_row_values import DatasetAppendRowValues

        d = dict(src_dict)
        values = DatasetAppendRowValues.from_dict(d.pop("values"))

        edit_type = cast(Union[Literal["append_row"], Unset], d.pop("edit_type", UNSET))
        if edit_type != "append_row" and not isinstance(edit_type, Unset):
            raise ValueError(f"edit_type must match const 'append_row', got '{edit_type}'")

        def _parse_row_id(data: object) -> Union[None, Unset, str]:
            if data is None:
                return data
            if isinstance(data, Unset):
                return data
            return cast(Union[None, Unset, str], data)

        row_id = _parse_row_id(d.pop("row_id", UNSET))

        dataset_append_row = cls(values=values, edit_type=edit_type, row_id=row_id)

        dataset_append_row.additional_properties = d
        return dataset_append_row

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from collections.abc import Mapping
from typing import Any, Literal, TypeVar, Union, cast

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET, Unset

T = TypeVar("T", bound="PromptTemplateVersionNumberSort")


@_attrs_define
class PromptTemplateVersionNumberSort:
    """
    Attributes:
        ascending (Union[Unset, bool]):  Default: True.
        name (Union[Literal['version'], Unset]):  Default: 'version'.
        sort_type (Union[Literal['column'], Unset]):  Default: 'column'.
    """

    ascending: Union[Unset, bool] = True
    name: Union[Literal["version"], Unset] = "version"
    sort_type: Union[Literal["column"], Unset] = "column"
    additional_properties: dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> dict[str, Any]:
        ascending = self.ascending

        name = self.name

        sort_type = self.sort_type

        field_dict: dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if ascending is not UNSET:
            field_dict["ascending"] = ascending
        if name is not UNSET:
            field_dict["name"] = name
        if sort_type is not UNSET:
            field_dict["sort_type"] = sort_type

        return field_dict

    @classmethod
    def from_dict(cls: type[T], src_dict: Mapping[str, Any]) -> T:
        d = dict(src_dict)
        ascending = d.pop("ascending", UNSET)

        name = cast(Union[Literal["version"], Unset], d.pop("name", UNSET))
        if name != "version" and not isinstance(name, Unset):
            raise ValueError(f"name must match const 'version', got '{name}'")

        sort_type = cast(Union[Literal["column"], Unset], d.pop("sort_type", UNSET))
        if sort_type != "column" and not isinstance(sort_type, Unset):
            raise ValueError(f"sort_type must match const 'column', got '{sort_type}'")

        prompt_template_version_number_sort = cls(ascending=ascending, name=name, sort_type=sort_type)

        prompt_template_version_number_sort.additional_properties = d
        return prompt_template_version_number_sort

    @property
    def additional_keys(self) -> list[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties
